from .layers import *
import numpy as np
from .tensor import *

class Optimizer:
    """
    Base class for optimizers used to update model parameters during training.

    Attributes
    ----------
    learning_rate : float
        The learning rate for the optimizer.
    momentum : float
        The momentum term for the optimizer.

    Methods
    -------
    update_params(layer : Layer)
        Updates the parameters of the given layer.
    """
    def __init__(self):
        """
        Initializes the Optimizer object.
        """
        self.learning_rate = None
        self.momentum = None

    def update_params(self, layer):
        """
        Updates the parameters of the given layer.

        Parameters
        ----------
        layer : Layer
            The layer whose parameters are to be updated.
        """
        pass

class SGD(Optimizer):
    """
    Stochastic Gradient Descent (SGD) optimizer.

    Methods
    -------
    update_params(layer : Layer)
        Updates the parameters of the given layer using SGD.
    """
    def __init__(self, learning_rate=0.001, momentum=None):
        """
        Initializes the SGD optimizer.
        """
        super().__init__()
        self.learning_rate = learning_rate
        self.momentum = momentum

        if not isinstance(learning_rate, float):
            raise TypeError("`learning_rate` must be a float")
        if not isinstance(momentum, float) and momentum is not None:
            raise TypeError("`momentum` must be a float or None")

    def update_params(self, layer):
        """
        Updates the parameters of the given layer using SGD.

        Parameters
        ----------
        layer : Layer
            The layer whose parameters are to be updated.

        Returns
        -------
        layer : Layer
            The layer with updated parameters.
        """
        if not isinstance(layer, Layer):
            raise TypeError("'layer' must be of type 'Layer'.")
        if isinstance(layer, Dense) or isinstance(layer, Convolution2D):
            if self.momentum is not None:
                weights = Tensor(self.momentum * layer.weight_momentums.array - layer.dW * self.learning_rate)
                if layer.biases is not None:
                    biases = Tensor(self.momentum * layer.bias_momentums.array - layer.dB * self.learning_rate)

                layer.weight_momentums = weights
                layer.bias_momentums = biases
                layer.weights += weights
                layer.biases += biases

            elif self.momentum is None:
                layer.weights -= Tensor(layer.dW * self.learning_rate)
                if layer.biases is not None:
                    layer.biases -= Tensor(layer.dB * self.learning_rate)

        return layer

class Adam(Optimizer):
    """
    Adam optimizer.

    Attributes
    ----------
    beta1 : float
        The exponential decay rate for the first moment estimates.
    beta2 : float
        The exponential decay rate for the second moment estimates.
    epsilon : float
        A small constant for numerical stability.
    step : int
        The current step of the optimizer.

    Methods
    -------
    update_params(layer : Layer)
        Updates the parameters of the given layer using Adam.
    """
    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999):
        """
        Initializes the Adam optimizer.
        """
        super().__init__()
        self.beta1 = beta1
        self.beta2 = beta2
        self.learning_rate = learning_rate
        self.epsilon = 1e-7
        self.step = 1

        if not isinstance(learning_rate, float):
            raise TypeError("`learning_rate` must be a float")
        if not isinstance(beta1, float):
            raise TypeError("`beta1` must be a float")
        if not isinstance(beta2, float):
            raise TypeError("`beta2` must be a float")

    def update_params(self, layer):
        """
        Updates the parameters of the given layer using Adam.

        Parameters
        ----------
        layer : Layer
            The layer whose parameters are to be updated.

        Returns
        -------
        layer : Layer
            The layer with updated parameters.
        """
        if not isinstance(layer, Layer):
            raise TypeError("'layer' must be of type 'Layer'.")
        if isinstance(layer, Dense) or isinstance(layer, Convolution2D):
            layer.weight_momentums = Tensor(self.beta1 * layer.weight_momentums.array + (1 - self.beta1) * layer.dW)
            weight_momentums_corrected = Tensor(layer.weight_momentums.array / (1 - self.beta1**self.step))

            layer.weight_cache = Tensor(self.beta2 * layer.weight_cache.array + (1 - self.beta2) * layer.dW**2)
            weight_cache_corrected = Tensor(layer.weight_cache.array / (1 - self.beta2**self.step))

            if layer.biases is not None:
                layer.bias_momentums = Tensor(self.beta1 * layer.bias_momentums.array + (1 - self.beta1) * layer.dB)
                bias_momentums_corrected = Tensor(layer.bias_momentums.array / (1 - self.beta1**self.step))

                layer.bias_cache = Tensor(self.beta2 * layer.bias_cache.array + (1 - self.beta2) * layer.dB**2)
                bias_cache_corrected = Tensor(layer.bias_cache.array / (1 - self.beta2**self.step))

            weights = Tensor(- self.learning_rate * weight_momentums_corrected.array / (np.sqrt(weight_cache_corrected.array) + self.epsilon))
            layer.weights += weights

            if layer.biases is not None:
                biases = Tensor(- self.learning_rate * bias_momentums_corrected.array / (np.sqrt(bias_cache_corrected.array) + self.epsilon))
                layer.biases += biases

            self.step += 1
        return layer


import numpy as np
from math import sqrt
from .im2col import *
from .tensor import *

class Layer:
    """
    Base class for all layers in the neural network.

    Attributes
    ----------
    inputs : Tensor
        Input data.
    outputs : Tensor
        Output data.
    dX : ndarray
        Partial derivative of loss with respect to input.
    dY : ndarray
        Partial derivative of loss with respect to output.

    Methods
    -------
    derivative()
        Computes the derivative of the layer.
    forward()
        Computes the forward pass of the layer.
    backward(dX : ndarray)
        Computes the backward pass of the layer.
    """
    def __init__(self):
        """
        Initializes the Layer class.
        """
        self.inputs = None
        self.outputs = None
        self.dX = None
        self.dY = None

    def derivative(self):
        """
        Computes the derivative of the layer.

        Returns
        -------
        ndarray
            The derivative of the layer.
        """
        pass

    def forward(self):
        """
        Computes the forward pass of the layer.

        Returns
        -------
        Tensor
            The output data.
        """
        pass

    def backward(self, dX):
        """
        Computes the backward pass of the layer.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        pass

    def zero_grad(self):
        pass

class Activation:
    """
    Base class for all activation functions.
    """
    def __init__(self):
        """
        Initializes the Activation class.
        """
        pass

class Dense(Layer):
    """
    A class that performs dense (fully connected) layer operations.

    Attributes
    ----------
    inputs : Tensor
        Input data.
    outputs : Tensor
        Output data.
    dX : ndarray
        Partial derivative of loss with respect to input.
    dY : ndarray
        Partial derivative of loss with respect to output.
    dW : ndarray
        Partial derivative of loss with respect to weight.
    dB : ndarray
        Partial derivative of loss with respect to bias.
    weights : Tensor
        Weights of the dense layer.
    biases : Tensor
        Biases of the dense layer.
    weight_momentums : Tensor
        Momentum for weights (also known as first order momentum).
    bias_momentums : Tensor
        Momentum for biases (also known as first order momentum).
    weight_cache : Tensor
        Cache for weights (also known as second order momentum).
    bias_cache : Tensor
        Cache for biases (also known as second order momentum).

    Methods
    -------
    forward()
        Computes the forward pass of the dense layer.
    backward(dX : ndarray)
        Computes the backward pass of the dense layer.
    """
    def __init__(self, n_in, n_out, weight_initializer="he_uniform", activation=None):
        """
        Initializes the Dense layer.

        Parameters
        ----------
        n_in : int
            Number of input features.
        n_out : int
            Number of output features.
        weight_initializer : str, optional
            Weight initialization method ('he_normal', 'glorot_normal', 'he_uniform', 'glorot_uniform').
            Default is 'he_uniform'.
        activation : Activation, optional
            Activation function ('relu', 'leaky relu', 'sigmoid', 'softmax').

        Raises
        ------
        ValueError
            If the weight initialization method is invalid.
        """
        super().__init__()

        def initialize_weights(weight_init, n_in, n_out):
            """
            Initializes the weights based on the specified method.

            Parameters
            ----------
            weight_init : str
                Weight initialization method.
            n_in : int
                Number of input features.
            n_out : int
                Number of output features.

            Returns
            -------
            Tensor
                Initialized weights.
            """
            if weight_init == "he_normal":
                weights = Tensor(np.random.randn(n_in, n_out) * np.sqrt(2.0 / n_in))
            elif weight_init == "he_uniform":
                limit = np.sqrt(6 / n_in)
                weights = Tensor(np.random.uniform(-limit, limit, (n_in, n_out)))
            elif weight_init == "glorot_normal":
                weights = Tensor(np.random.randn(n_in, n_out) * np.sqrt(2.0 / (n_in + n_out)))
            elif weight_init == "glorot_uniform":
                limit = np.sqrt(6 / (n_in + n_out))
                weights = Tensor(np.random.uniform(-limit, limit, (n_in, n_out)))
            else:
                raise ValueError(
                    "`weight_init` must be either 'he_uniform', 'he_normal', 'glorot_uniform' or 'glorot_normal'.")
            return weights

        self.weights = initialize_weights(weight_initializer, n_in, n_out)
        self.biases = Tensor(np.random.rand(1, n_out))
        self.dW = np.zeros_like(self.weights)
        self.dB = np.zeros_like(self.biases)
        self.weight_momentums = Tensor(np.zeros_like(self.weights))
        self.bias_momentums = Tensor(np.zeros_like(self.biases))
        self.weight_cache = Tensor(np.zeros_like(self.weights))
        self.bias_cache = Tensor(np.zeros_like(self.biases))
        self.activation = activation

        if not isinstance(activation, Activation) and activation is not None:
            raise TypeError("`activation` must be of type `Activation`.")

    def forward(self):
        """
        Computes the forward pass of the dense layer.

        Returns
        -------
        Tensor
            The output data.
        """
        self.outputs = dot(self.inputs, self.weights) + self.biases

        if self.activation is not None:
            self.activation.inputs = self.outputs
            return self.activation.forward()

        return self.outputs

    def backward(self, dX):
        """
        Computes the backward pass of the dense layer.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        if self.activation is not None:
            dX = self.activation.backward(dX)

        self.dY = dX
        self.outputs.backward(self.dY)
        self.dX = self.inputs.grad
        self.dW = self.weights.grad
        return self.dX

    def zero_grad(self):
        if self.activation is not None:
            self.activation.zero_grad()
        self.outputs.zero_grad()

class ReLU(Layer, Activation):
    """
    A class that performs ReLU (Rectified Linear Unit) layer operations.

    Attributes
    ----------
    inputs : Tensor
        Input data.
    outputs : Tensor
        Output data.
    dX : ndarray
        Partial derivative of loss with respect to input.
    dY : ndarray
        Partial derivative of loss with respect to output.

    Methods
    -------
    derivative()
        Computes the ReLU derivative.
    forward()
        Computes the ReLU forward pass.
    backward(dX : ndarray)
        Computes the ReLU backward pass.
    """
    def __init__(self):
        """
        Initializes the ReLU layer.
        """
        super().__init__()

    def derivative(self):
        """
        Computes the ReLU derivative.

        Returns
        -------
        ndarray
            The differentiated input data.
        """
        dA = np.ones_like(self.inputs.array)
        dA[self.inputs.array <= 0] = 0
        return dA

    def forward(self):
        """
        Computes the ReLU forward pass.

        Returns
        -------
        Tensor
            The output data.
        """
        self.outputs = Tensor(np.maximum(0, self.inputs.array))
        return self.outputs

    def backward(self, dX):
        """
        Computes the ReLU backward pass.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        self.dY = dX
        self.dX = self.dY * self.derivative()
        return self.dX

class LeakyReLU(Layer, Activation):
    """
    A class that performs Leaky ReLU layer operations.

    Attributes
    ----------
    inputs : Tensor
        Input data.
    outputs : Tensor
        Output data.
    dX : ndarray
        Partial derivative of loss with respect to input.
    dY : ndarray
        Partial derivative of loss with respect to output.

    Methods
    -------
    derivative()
        Computes the Leaky ReLU derivative.
    forward()
        Computes the Leaky ReLU forward pass.
    backward(dX : ndarray)
        Computes the Leaky ReLU backward pass.
    """
    def __init__(self):
        """
        Initializes the LeakyReLU layer.
        """
        super().__init__()

    def derivative(self):
        """
        Computes the Leaky ReLU derivative.

        Returns
        -------
        ndarray
            The differentiated input data with respect to loss.
        """
        dA = np.ones_like(self.inputs.array)
        dA[self.inputs.array <= 0] = 0.01
        return dA

    def forward(self):
        """
        Computes the Leaky ReLU forward pass.

        Returns
        -------
        Tensor
            The output data.
        """
        self.outputs = Tensor(np.where(self.inputs.array > 0, self.inputs.array, self.inputs.array * 0.01))
        return self.outputs

    def backward(self, dX):
        """
        Computes the Leaky ReLU backward pass.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        self.dY = dX
        self.dX = self.dY * self.derivative()
        return self.dX

class Sigmoid(Layer, Activation):
    """
    A class that performs Sigmoid layer operations.

    Attributes
    ----------
    inputs : Tensor
        Input data.
    outputs : Tensor
        Output data.
    dX : ndarray
        Partial derivative of loss with respect to input.
    dY : ndarray
        Partial derivative of loss with respect to output.

    Methods
    -------
    derivative()
        Computes the Sigmoid derivative.
    forward()
        Computes the Sigmoid forward pass.
    backward(dX : ndarray)
        Computes the Sigmoid backward pass.
    """
    def __init__(self):
        """
        Initializes the Sigmoid layer.
        """
        super().__init__()

    def derivative(self):
        """
        Computes the Sigmoid derivative.

        Returns
        -------
        ndarray
            The Sigmoid derivative.
        """
        return self.outputs * (1 - self.outputs)

    def forward(self):
        """
        Computes the Sigmoid forward pass.

        Returns
        -------
        Tensor
            The output data.
        """
        self.outputs = 1 / (1 + exp(-self.inputs))
        return self.outputs

    def backward(self, dX):
        """
        Computes the Sigmoid backward pass.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        self.dY = dX
        self.outputs.backward(self.dY)
        self.dX = self.inputs.grad
        return self.dX

    def zero_grad(self):
        self.outputs.zero_grad()

class Softmax(Layer, Activation):
    """
    A class that performs Softmax layer operations.

    Attributes
    ----------
    inputs : Tensor
        Input data.
    outputs : Tensor
        Output data.
    dX : ndarray
        Partial derivative of loss with respect to input.
    dY : ndarray
        Partial derivative of loss with respect to output.

    Methods
    -------
    forward()
        Computes the Softmax forward pass.
    backward(dX : ndarray)
        Computes the Softmax backward pass.
    """
    def __init__(self):
        """
        Initializes the Softmax layer.
        """
        super().__init__()

    def forward(self):
        """
        Computes the Softmax forward pass.

        Returns
        -------
        Tensor
            The output data.
        """
        exp_values = exp(self.inputs - max(self.inputs, axis=1, keepdims=True))
        probabilities = exp_values / sum(exp_values, axis=1, keepdims=True)
        self.outputs = probabilities
        return self.outputs

    def backward(self, dX):
        """
        Computes the Softmax backward pass.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        self.dY = dX
        self.outputs.backward(self.dY)
        self.dX = self.inputs.grad
        return self.dX

    def zero_grad(self):
        self.outputs.zero_grad()

class Convolution2D(Layer):
    """
    A class that performs 2D convolution layer operations.

    Attributes
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    weights : Tensor
        Filters of the convolution layer.
    padding : str
        Padding type ('valid' or 'same').
    kernel_size : int
        Size of the convolution kernel.
    stride : int
        Stride of the convolution.
    weight_momentums : Tensor
        Momentum for weights.
    weight_cache : Tensor
        Cache for weights.
    biases : Tensor
        Biases of the convolution layer.
    bias_momentums : Tensor
        Momentum for biases.
    bias_cache : Tensor
        Cache for biases.

    Methods
    -------
    calculate_padding()
        Calculates the padding for the input.
    explicit_padding()
        Applies explicit padding to the input.
    get_output_size(input_height : int, input_width : int)
        Gets the output size of the convolution.
    forward()
        Computes the forward pass of the convolution layer.
    backward(dX : ndarray)
        Computes the backward pass of the convolution layer.
    """
    def __init__(self, in_channels, out_channels, kernel_size, padding="valid", stride=1, weight_initializer="he_uniform", use_bias=True, activation=None):
        """
        Initializes the Convolution2D layer.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : int
            Size of the convolution kernel.
        padding : str, optional
            Padding type ('valid' or 'same'). Default is 'valid'.
        stride : int, optional
            Stride of the convolution. Default is 1.
        weight_initializer : str, optional
            Weight initialization method ('he_normal', 'glorot_normal', 'he_uniform', 'glorot_uniform').
            Default is 'he_uniform'.
        activation : Activation, optional
            Activation function. Default is None.

        Raises
        ------
        TypeError
            If `padding` is not a string.
        ValueError
            If `padding` is not 'valid' or 'same'.
        """
        super().__init__()

        def initialize_weights(weight_init, in_channels, out_channels, kernel_size):
            """
            Initializes the weights based on the specified method.

            Parameters
            ----------
            weight_init : str
                Weight initialization method.
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            kernel_size : int
                Size of the convolution kernel.

            Returns
            -------
            Tensor
                Initialized weights.
            """
            if weight_init == "he_normal":
                weights = Tensor(np.random.randn(out_channels, in_channels, kernel_size, kernel_size) * np.sqrt(
                    2.0 / (in_channels * kernel_size ** 2)))
            elif weight_init == "he_uniform":
                limit = np.sqrt(6 / (in_channels * kernel_size ** 2))
                weights = Tensor(
                    np.random.uniform(-limit, limit, (out_channels, in_channels, kernel_size, kernel_size)))
            elif weight_init == "glorot_normal":
                weights = Tensor(np.random.randn(out_channels, in_channels, kernel_size, kernel_size) * np.sqrt(
                    2.0 / (in_channels * kernel_size ** 2 + out_channels * kernel_size ** 2)))
            elif weight_init == "glorot_uniform":
                limit = np.sqrt(6 / (in_channels * kernel_size ** 2 + out_channels * kernel_size ** 2))
                weights = Tensor(
                    np.random.uniform(-limit, limit, (out_channels, in_channels, kernel_size, kernel_size)))
            else:
                raise ValueError(
                    "`weight_init` must be either 'he_uniform', 'he_normal', 'glorot_uniform' or 'glorot_normal'.")
            return weights

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.weights = initialize_weights(weight_initializer, in_channels, out_channels, kernel_size)
        self.dW = np.zeros_like(self.weights)
        self.weight_momentums = Tensor(np.zeros_like(self.weights))
        self.weight_cache = Tensor(np.zeros_like(self.weights))
        if use_bias is True:
            self.biases = Tensor(np.zeros(shape=(1, out_channels, 1, 1)).astype(np.float64))
            self.dB = np.zeros_like(self.biases)
            self.bias_momentums = Tensor(np.zeros_like(self.biases))
            self.bias_cache = Tensor(np.zeros_like(self.biases))
        else:
            self.biases = None
        self.padding = padding
        self.kernel_size = kernel_size
        self.stride = stride
        self.activation = activation

        if not isinstance(activation, Activation) and activation is not None:
            raise TypeError("`activation` must be of type `Activation`.")

        if not isinstance(padding, str):
            raise TypeError("`padding` must be a string.")
        if self.padding not in ("valid", "same"):
            raise ValueError("`padding` must be 'valid' or 'same'.")

        if self.padding == "same":
            self.stride = 1

    def calculate_padding(self):
        """
        Calculates the padding for the input.

        Returns
        -------
        tuple
            Padding values (top, bottom, left, right).
        """
        if self.padding == "valid":
            return (0, 0, 0, 0)
        elif self.padding == "same":
            input_height, input_width = self.inputs.shape[2], self.inputs.shape[3]
            if input_height % self.stride == 0:
                pad_along_height = max((self.kernel_size - self.stride), 0)
            else:
                pad_along_height = max(self.kernel_size - (input_height % self.stride), 0)
            if input_width % self.stride == 0:
                pad_along_width = max((self.kernel_size - self.stride), 0)
            else:
                pad_along_width = max(self.kernel_size - (input_width % self.stride), 0)

            pad_top = pad_along_height.array.astype(int) // 2
            pad_bottom = pad_along_height.array.astype(int) - pad_top
            pad_left = pad_along_width.array.astype(int) // 2
            pad_right = pad_along_width.array.astype(int) - pad_left

            return (pad_top, pad_bottom, pad_left, pad_right)

    def explicit_padding(self):
        """
        Applies explicit padding to the input.

        Returns
        -------
        Tensor
            Padded input data.
        """
        pad_top, pad_bottom, pad_left, pad_right = self.calculate_padding()
        return Tensor(np.pad(self.inputs.array, ((0, 0), (0, 0), (pad_top, pad_bottom), (pad_left, pad_right)), mode='constant'))

    def get_output_size(self, input_height, input_width):
        """
        Gets the output size of the convolution.

        Parameters
        ----------
        input_height : int
            Height of the input.
        input_width : int
            Width of the input.

        Returns
        -------
        tuple
            Output height and width.
        """
        if self.padding == 'valid':
            output_height = (input_height - self.kernel_size) // self.stride + 1
            output_width = (input_width - self.kernel_size) // self.stride + 1
        elif self.padding == 'same':
            output_height = np.ceil(input_height / self.stride)
            output_width = np.ceil(input_width / self.stride)

        return int(output_height), int(output_width)

    def forward(self):
        """
        Computes the forward pass of the convolution layer.

        Returns
        -------
        Tensor
            The output data.
        """
        self.input_padded = self.explicit_padding()

        self.input_cols = im2col(self.input_padded, self.kernel_size, self.stride)
        self.filter_cols = Tensor(self.weights.array.reshape(self.out_channels, -1))

        output_height, output_width = self.get_output_size(self.inputs.shape[2], self.inputs.shape[3])

        self.output_cols = self.filter_cols @ self.input_cols

        self.outputs = Tensor(np.array(np.hsplit(self.output_cols.array, self.inputs.shape[0])).reshape(
            (self.input_padded.shape[0], self.out_channels, output_height, output_width)
        ))

        if self.biases is not None:
            self.outputs = self.outputs + self.biases

        if self.activation is not None:
            self.activation.inputs = self.outputs
            return self.activation.forward()

        return self.outputs

    def backward(self, dX):
        """
        Computes the backward pass of the convolution layer.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        if self.activation is not None:
            dX = self.activation.backward(dX)

        self.dY = dX

        self.dY_reshaped = self.dY.reshape(self.dY.shape[0] * self.dY.shape[1], self.dY.shape[2] * self.dY.shape[3])
        self.dY_reshaped = np.array(np.vsplit(self.dY_reshaped, self.inputs.shape[0]))
        self.dY_reshaped = np.concatenate(self.dY_reshaped, axis=-1)

        self.output_cols.backward(self.dY_reshaped)

        self.dX_cols = self.input_cols.grad
        self.dW_cols = self.filter_cols.grad

        self.dX_padded = col2im(Tensor(self.dX_cols), self.input_padded.shape, self.kernel_size, self.stride)

        if self.padding == "same":
            (pad_top, pad_bottom, pad_left, pad_right) = self.calculate_padding()
            self.dX = self.dX_padded.array[:, :, pad_top:-pad_bottom, pad_left:-pad_right]
        else:
            self.dX = self.dX_padded.array

        self.dW = self.dW_cols.reshape((self.dW_cols.shape[0], self.in_channels, self.kernel_size, self.kernel_size))

        self.outputs.backward(self.dY)
        self.dB = self.biases.grad

        return self.dX

    def zero_grad(self):
        if self.activation is not None:
            self.activation.zero_grad()
        self.output_cols.zero_grad()

class Pooling2D(Layer):
    """
    A class that performs 2D pooling layer operations.

    Attributes
    ----------
    pool_size : int
        Size of the pooling window.
    stride : int
        Stride of the pooling.
    mode : str
        Pooling mode ('max').

    Methods
    -------
    forward()
        Computes the forward pass of the pooling layer.
    backward(dX : ndarray)
        Computes the backward pass of the pooling layer.
    """
    def __init__(self, pool_size, stride, mode="max"):
        """
        Initializes the Pooling2D layer.

        Parameters
        ----------
        pool_size : int
            Size of the pooling window.
        stride : int
            Stride of the pooling.
        mode : str, optional
            Pooling mode ('max'). Default is 'max'.

        Raises
        ------
        ValueError
            If the pooling mode is invalid.
        """
        super().__init__()
        self.pool_size = pool_size
        self.stride = stride
        self.mode = mode

        if self.mode not in ("max"):
            raise ValueError("`mode` must be 'max'.")

    def forward(self):
        """
        Computes the forward pass of the pooling layer.

        Returns
        -------
        Tensor
            The output data.
        """
        output_height = int((self.inputs.shape[2] - self.pool_size + self.stride) // self.stride)
        output_width = int((self.inputs.shape[3] - self.pool_size + self.stride) // self.stride)

        output_shape = (self.inputs.shape[0], self.inputs.shape[1], output_height, output_width)

        self.input_cols = im2col(self.inputs, self.pool_size, self.stride)
        self.input_cols_reshaped = Tensor(
            np.array(np.hsplit(np.array(np.hsplit(self.input_cols.array, self.inputs.shape[0])), self.inputs.shape[1])))

        self.maxima = max(self.input_cols_reshaped, axis=2)
        self.maxima_reshaped = self.maxima.array.reshape(self.inputs.shape[1], -1)

        self.outputs = col2im(Tensor(self.maxima_reshaped), output_shape, 1, 1)

        return self.outputs

    def backward(self, dX):
        """
        Computes the backward pass of the pooling layer.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        self.dY = dX
        self.dX = np.zeros_like(self.inputs)

        self.dY_cols = im2col(Tensor(self.dY), 1, 1)
        self.dY_cols_reshaped = np.array(
            np.hsplit(np.array(np.hsplit(self.dY_cols.array, self.dY.shape[0])), self.dY.shape[1]))

        self.maxima.backward(self.dY_cols_reshaped)

        self.dX_cols = np.concatenate(np.concatenate(self.input_cols_reshaped.grad, axis=1), axis=1)

        self.dX = col2im(Tensor(self.dX_cols), self.inputs.shape, self.pool_size, self.stride).array

        return self.dX

    def zero_grad(self):
        self.maxima.zero_grad()

class Flatten(Layer):
    """
    A class that performs flattening layer operations.

    Methods
    -------
    forward()
        Computes the forward pass of the flattening layer.
    backward(dX : ndarray)
        Computes the backward pass of the flattening layer.
    """
    def __init__(self):
        """
        Initializes the Flatten layer.
        """
        super().__init__()

    def forward(self):
        """
        Computes the forward pass of the flattening layer.

        Returns
        -------
        Tensor
            The output data.
        """
        self.outputs = Tensor(self.inputs.array.reshape((self.inputs.shape[0], -1)))
        return self.outputs

    def backward(self, dX):
        """
        Computes the backward pass of the flattening layer.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        self.dY = dX
        self.dX = self.dY.reshape(self.inputs.shape)
        return self.dX

class Dropout(Layer):
    """
    A class that performs dropout layer operations.

    Attributes
    ----------
    p : float
        Dropout probability.
    mask : ndarray
        Dropout mask.
    training : bool
        Whether the layer is in training mode.

    Methods
    -------
    forward()
        Computes the forward pass of the dropout layer.
    backward(dX : ndarray)
        Computes the backward pass of the dropout layer.
    """
    def __init__(self, p):
        """
        Initializes the Dropout layer.

        Parameters
        ----------
        p : float
            Dropout probability. Must be between 0 and 1.

        Raises
        ------
        ValueError
            If the dropout probability is not between 0 and 1.
        """
        super().__init__()
        if p < 0 or p > 1:
            raise ValueError("Dropout probability has to be between 0 and 1")
        self.p = p
        self.mask = None
        self.training = True

    def forward(self):
        """
        Computes the forward pass of the dropout layer.

        Returns
        -------
        Tensor
            The output data.
        """
        if self.training:
            self.mask = np.random.binomial(1, 1 - self.p, size=self.inputs.shape)
            self.outputs = Tensor(self.inputs.array * self.mask * 1.0 / (1.0 - self.p))
        else:
            self.outputs = self.inputs
        return self.outputs

    def backward(self, dX):
        """
        Computes the backward pass of the dropout layer.

        Parameters
        ----------
        dX : ndarray
            Partial derivative of loss with respect to output data.

        Returns
        -------
        ndarray
            Partial derivative of loss with respect to input data.
        """
        if self.training:
            self.dX = dX * self.mask * 1.0 / (1.0 - self.p)
        else:
            self.dX = dX
        return self.dX

class Linear:
    """
    A class that performs linear layer operations.

    Attributes
    ----------
    weights : ndarray
        Weights of the linear layer.
    biases : ndarray
        Biases of the linear layer.

    Methods
    -------
    forward()
        Computes the forward pass of the linear function.
    backward(lr : float)
        Computes the backward pass of the linear layer.
    """
    def __init__(self, n_in, n_out):
        """
        Initializes the Linear layer.

        Parameters
        ----------
        n_in : int
            Number of input features.
        n_out : int
            Number of output features.
        """
        self.weights = np.random.rand(n_in, n_out)
        self.biases = np.random.rand(1, n_out)
        self.dW = None
        self.dB = None

    def forward(self):
        """
        Computes the forward pass of the linear function.

        Returns
        -------
        ndarray
            The output data.
        """
        self.outputs = np.dot(self.inputs, self.weights) + self.biases
        return self.outputs

    def backward(self, lr):
        """
        Computes the backward pass of the linear function.

        Parameters
        ----------
        lr : float
            Learning rate.
        """
        self.dW = np.sum(-2 * self.inputs * (self.targets - (self.inputs @ self.weights + self.biases))) / len(self.targets - self.outputs)
        self.dB = np.sum(-2 * (self.targets - (self.inputs @ self.weights + self.biases))) / len(self.targets - self.outputs)
        self.weights -= self.dW * lr
        self.biases -= self.dB * lr
        self.dW *= np.zeros(self.dW.shape, dtype=np.float64)
        self.dB *= np.zeros(self.dB.shape, dtype=np.float64)



import numpy as np

class RobustScaler:
    """
    A class for scaling features using statistics that are robust to outliers.

    Attributes
    ----------
    fitted : bool
        Indicates whether the scaler has been fitted to the data.
    q1_list : list
        List of the first quartile (25th percentile) values for each feature.
    median_list : list
        List of the median (50th percentile) values for each feature.
    q3_list : list
        List of the third quartile (75th percentile) values for each feature.

    Methods
    -------
    fit(X : ndarray)
        Computes the quartiles and median for scaling.
    transform(X : ndarray)
        Scales the data using the computed quartiles and median.
    """
    def __init__(self):
        """
        Initializes the RobustScaler object.
        """
        self.fitted = False
        self.q1_list, self.median_list, self.q3_list = [], [], []

    def fit(self, X):
        """
        Computes the quartiles and median for scaling.

        Parameters
        ----------
        X : ndarray
            The input data to fit the scaler.

        Raises
        ------
        ValueError
            If the input data has more than 2 dimensions.
        """
        if X.ndim > 2:
            raise ValueError("scalers don't support more than 2 dimensions.")
        self.q1_list = np.quantile(X, 0.25, axis=0)
        self.median_list = np.quantile(X, 0.5, axis=0)
        self.q3_list = np.quantile(X, 0.75, axis=0)
        self.fitted = True

    def transform(self, X):
        """
        Scales the data using the computed quartiles and median.

        Parameters
        ----------
        X : ndarray
            The input data to be scaled.

        Returns
        -------
        X_scaled : ndarray
            The scaled data.
        """
        if not self.fitted:
            self.fit(X)
        X_scaled = (X - self.median_list) / (self.q3_list - self.q1_list + 1e-5)
        return X_scaled

class MinMaxScaler:
    """
    A class for scaling features to a given range.

    Attributes
    ----------
    fitted : bool
        Indicates whether the scaler has been fitted to the data.
    min_list : list
        List of the minimum values for each feature.
    max_list : list
        List of the maximum values for each feature.

    Methods
    -------
    fit(X : ndarray)
        Computes the minimum and maximum values for scaling.
    transform(X : ndarray)
        Scales the data using the computed minimum and maximum values.
    """
    def __init__(self):
        """
        Initializes the MinMaxScaler object.
        """
        self.fitted = False
        self.min_list, self.max_list = [], []

    def fit(self, X):
        """
        Computes the minimum and maximum values for scaling.

        Parameters
        ----------
        X : ndarray
            The input data to fit the scaler.

        Raises
        ------
        ValueError
            If the input data has more than 2 dimensions.
        """
        if X.ndim > 2:
            raise ValueError("scalers don't support more than 2 dimensions")
        self.min_list = np.min(X, axis=0)
        self.max_list = np.max(X, axis=0)
        self.fitted = True

    def transform(self, X):
        """
        Scales the data using the computed minimum and maximum values.

        Parameters
        ----------
        X : ndarray
            The input data to be scaled.

        Returns
        -------
        X_scaled : ndarray
            The scaled data.
        """
        if not self.fitted:
            self.fit(X)
        X_scaled = (X - self.min_list) / (self.max_list - self.min_list + 1e-5)
        return X_scaled

class StandardScaler:
    """
    A class for standardizing features by removing the mean and scaling to unit variance.

    Attributes
    ----------
    fitted : bool
        Indicates whether the scaler has been fitted to the data.
    mean_list : list
        List of the mean values for each feature.
    std_list : list
        List of the standard deviation values for each feature.

    Methods
    -------
    fit(X : ndarray)
        Computes the mean and standard deviation for scaling.
    transform(X : ndarray)
        Scales the data using the computed mean and standard deviation.
    """
    def __init__(self):
        """
        Initializes the StandardScaler object.
        """
        self.fitted = False
        self.mean_list, self.std_list = [], []

    def fit(self, X):
        """
        Computes the mean and standard deviation for scaling.

        Parameters
        ----------
        X : ndarray
            The input data to fit the scaler.

        Raises
        ------
        ValueError
            If the input data has more than 2 dimensions.
        """
        if X.ndim > 2:
            raise ValueError("scalers don't support more than 2 dimensions")
        self.mean_list = np.mean(X, axis=0)
        self.std_list = np.std(np.array(X, dtype=np.float64), axis=0)
        self.fitted = True

    def transform(self, X):
        """
        Scales the data using the computed mean and standard deviation.

        Parameters
        ----------
        X : ndarray
            The input data to be scaled.

        Returns
        -------
        X_scaled : ndarray
            The scaled data.
        """
        if not self.fitted:
            self.fit(X)
        X_scaled = (X - self.mean_list) / (self.std_list + 1e-5)
        return X_scaled

class OneHotEncoder:
    """
    A class for encoding categorical features as a one-hot numeric array.

    Attributes
    ----------
    fitted : bool
        Indicates whether the encoder has been fitted to the data.
    categories : list
        List of unique categories found in the data.

    Methods
    -------
    fit(y : ndarray)
        Fits the encoder to the data by identifying unique categories.
    transform(y : ndarray)
        Transforms the data into a one-hot encoded array.
    """
    def __init__(self):
        """
        Initializes the OneHotEncoder object.
        """
        self.fitted = False
        self.categories = []

    def fit(self, y):
        """
        Fits the encoder to the data by identifying unique categories.

        Parameters
        ----------
        y : ndarray
            The input data to fit the encoder.
        """
        y = np.asarray(y).flatten()
        self.categories = np.unique(y)
        self.fitted = True

    def transform(self, y):
        """
        Transforms the data into a one-hot encoded array.

        Parameters
        ----------
        y : ndarray
            The input data to be transformed.

        Returns
        -------
        result : ndarray
            The one-hot encoded array.
        """
        if not self.fitted:
            self.fit(y)
        y = np.asarray(y).flatten()
        result = np.zeros((len(y), len(self.categories)))
        for i, category in enumerate(self.categories):
            result[:, i] = (y == category)
        return np.array(result, dtype=np.float64)

class StringEncoder:
    """
    A class for encoding string features using one-hot encoding.

    Attributes
    ----------
    feature_indices : list
        List of indices of string features.
    encoders : list
        List of OneHotEncoder objects for each string feature.
    fitted : bool
        Indicates whether the encoder has been fitted to the data.

    Methods
    -------
    transform(X : ndarray)
        Transforms the data by encoding string features.
    fit(X : ndarray)
        Fits the encoder to the data by identifying string features and fitting OneHotEncoder objects.
    _encode(X : ndarray)
        Encodes the string features using the fitted OneHotEncoder objects.
    """
    def __init__(self):
        """
        Initializes the StringEncoder object.
        """
        self.feature_indices = []
        self.encoders = []
        self.fitted = False

    def transform(self, X):
        """
        Transforms the data by encoding string features.

        Parameters
        ----------
        X : ndarray
            The input data to be transformed.

        Returns
        -------
        X_encoded : ndarray
            The transformed data with string features encoded.
        """
        if not self.fitted:
            self.fit(X)
        return self._encode(X)

    def fit(self, X):
        """
        Fits the encoder to the data by identifying string features and fitting OneHotEncoder objects.

        Parameters
        ----------
        X : ndarray
            The input data to fit the encoder.
        """
        num_features = X.shape[1]
        for i in range(num_features):
            if isinstance(X[0, i], str):
                encoder = OneHotEncoder()
                encoder.fit(X[:, i])
                self.feature_indices.append(i)
                self.encoders.append(encoder)
        self.fitted = True

    def _encode(self, X):
        """
        Encodes the string features using the fitted OneHotEncoder objects.

        Parameters
        ----------
        X : ndarray
            The input data to be encoded.

        Returns
        -------
        X_encoded : ndarray
            The encoded data.
        """
        for i, encoder in zip(self.feature_indices, self.encoders):
            transformed = encoder.transform(X[:, i])
            X = np.delete(X, i, axis=1)
            X = np.hstack((X[:, :i], transformed, X[:, i:]))
        return np.array(X, dtype=np.float64)

class SimpleImputer:
    """
    A class for imputing missing values in a dataset.

    Methods
    -------
    transform(X, column, missing_values=np.nan, strategy="mean")
        Imputes missing values in the specified column using the specified strategy.
    """
    def transform(self, X, column, missing_values=np.nan, strategy="mean"):
        """
        Imputes missing values in the specified column using the specified strategy.

        Parameters
        ----------
        X : ndarray
            The input data.
        column : int
            The index of the column to impute.
        missing_values : float or str, optional
            The value to consider as missing. Default is np.nan.
        strategy : str, optional
            The imputation strategy. Currently, only "mean" is supported. Default is "mean".

        Returns
        -------
        X_copy : ndarray
            The input data with missing values imputed.

        Raises
        ------
        ValueError
            If the strategy is not one of ("mean").
        TypeError
            If the column data type is not numeric.
        """
        X_copy = np.copy(X)

        if strategy != "mean":
            raise ValueError("`strategy` must be one of ('mean').")

        if X_copy[:, column].dtype == "<U4":
            raise TypeError("`column` must be numeric.")

        if np.isnan(missing_values):
            mask = np.isnan(X_copy[:, column].astype(float))
            mean = np.nanmean(X_copy[:, column].astype(float))
        else:
            mask = (X_copy[:, column] == missing_values)
            filtered_values = X_copy[~mask, column].astype(float)
            mean = np.mean(filtered_values) if filtered_values.size > 0 else 0

        X_copy[mask, column] = mean
        return X_copy



import numpy as np

class Tensor:
    def __init__(self, object, requires_grad=True, *args, **kwargs):
        self.array = np.array(object, *args, **kwargs).astype(np.float64)
        self.requires_grad = requires_grad
        self.grad = np.zeros_like(self.array)

    @property
    def shape(self):
        return self.array.shape

    @property
    def size(self):
        return self.array.size

    @property
    def ndim(self):
        return self.array.ndim

    @property
    def T(self):
        return Tensor(self.array.T)

    def __str__(self):
        return str(self.array)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.array})"

    def __add__(self, value):
        return add(self, value)

    def __radd__(self, value):
        return self.__add__(value)

    def __iadd__(self, value):
        self.array += value.array if isinstance(value, Tensor) else value
        return self

    def __sub__(self, value):
        return subtract(self, value)

    def __rsub__(self, value):
        return subtract(value, self)

    def __isub__(self, value):
        self.array -= value.array if isinstance(value, Tensor) else value
        return self

    def __neg__(self):
        return subtract(0, self)

    def __mul__(self, value):
        return multiply(self, value)

    def __rmul__(self, value):
        return self.__mul__(value)

    def __imul__(self, value):
        self.array *= value.array if isinstance(value, Tensor) else value
        return self

    def __truediv__(self, value):
        return divide(self.array, value)

    def __rtruediv__(self, value):
        return divide(value, self)

    def __itruediv__(self, value):
        self.array /= value.array if isinstance(value, Tensor) else value
        return self

    def __floordiv__(self, value):
        return floor_divide(self, value)

    def __matmul__(self, value):
        return matmul(self, value)

    def __rmatmul__(self, value):
        return self.__matmul__(value)

    def __imatmul__(self, value):
        self.array @= value.array if isinstance(value, Tensor) else value
        return self

    def __pow__(self, value):
        return power(self, value)

    def __ipow__(self, value):
        self.array **= value.array if isinstance(value, Tensor) else value
        return self

    def __eq__(self, value):
        return Tensor(self.array == (value.array if isinstance(value, Tensor) else value))

    def __ne__(self, value):
        return Tensor(self.array != (value.array if isinstance(value, Tensor) else value))

    def __gt__(self, value):
        return Tensor(self.array > (value.array if isinstance(value, Tensor) else value))

    def __ge__(self, value):
        return Tensor(self.array >= (value.array if isinstance(value, Tensor) else value))

    def __lt__(self, value):
        return Tensor(self.array < (value.array if isinstance(value, Tensor) else value))

    def __le__(self, value):
        return Tensor(self.array <= (value.array if isinstance(value, Tensor) else value))

    def __len__(self):
        return len(self.array)

    def zero_grad(self):
        self.grad = np.zeros_like(self.array)

    def to_numpy(self):
        return self.array

class zeros(Tensor):
    def __init__(self, *args, **kwargs):
        array = np.zeros(*args, **kwargs)
        super().__init__(object = array)

class zeros_like(Tensor):
    def __init__(self, a,  *args, **kwargs):
        def _get_array(obj):
            if isinstance(obj, Operation):
                return obj.output.array
            elif isinstance(obj, Tensor):
                return obj.array
            else:
                return np.array(obj)

        array = np.zeros_like(_get_array(a), *args, **kwargs)
        super().__init__(object = array)

class Operation:
    def __init__(self, x1, x2=None, *args, **kwargs):
        self.x1 = x1
        self.x2 = x2
        self.output = None
        self.__dict__.update(kwargs)
        self.forward(*args, **kwargs)

    @property
    def grad(self):
        if self.output is not None:
            return self.output.grad

    @property
    def array(self):
        if self.output is not None:
            return self.output.array

    @property
    def shape(self):
        if self.output is not None:
            return self.output.shape

    @property
    def size(self):
        if self.output is not None:
            return self.output.size

    @property
    def T(self):
        if self.output is not None:
            return self.output.T

    @property
    def ndim(self):
        if self.output is not None:
            return self.output.ndim

    def __str__(self):
        return self.output.array.__str__()

    def __repr__(self):
        return self.__str__()

    def __neg__(self):
        return subtract(0, self)

    def __add__(self, value):
        return add(self, value)

    def __radd__(self, value):
        return self.__add__(value)

    def __sub__(self, value):
        return subtract(self, value)

    def __rsub__(self, value):
        return subtract(value, self)

    def __mul__(self, value):
        return multiply(self, value)

    def __rmul__(self, value):
        return self.__mul__(value)

    def __truediv__(self, value):
        return divide(self, value)

    def __rtruediv__(self, value):
        return divide(value, self)

    def __floordiv__(self, value):
        return floor_divide(self, value)

    def __matmul__(self, value):
        return matmul(self, value)

    def __rmatmul__(self, value):
        return self.__matmul__(value)

    def __pow__(self, value):
        return power(self, value).output

    def forward(self, *args, **kwargs):
        pass

    def backward(self, grad):
        pass

    def zero_grad(self):
        if isinstance(self.x1, Operation) or isinstance(self.x1, Tensor):
            self.x1.zero_grad()
        if hasattr(self, "x2") and (isinstance(self.x2, Operation) or isinstance(self.x2, Tensor)):
            self.x2.zero_grad()
        if self.output is not None:
            self.output.zero_grad()

    def _get_array(self, obj):
        if isinstance(obj, Operation):
            return obj.output.array
        elif isinstance(obj, Tensor):
            return obj.array
        else:
            return np.array(obj)

    def _apply_grad(self, obj, grad):
        if isinstance(obj, Operation):
            obj.output.grad += grad
            obj.backward(grad)
        elif isinstance(obj, Tensor) and obj.requires_grad:
            obj.grad += grad

    def _requires_grad(self, *inputs):
        return any(isinstance(i, Tensor) and i.requires_grad for i in inputs)

    def _compress_grad(self, grad, tensor):
        tensor = self._get_array(tensor)
        extra_dims = np.array(grad).ndim - tensor.ndim
        for _ in range(extra_dims):
            grad = np.sum(grad, axis=0)
        for i, dim in enumerate(tensor.shape):
            if dim == 1:
                grad = np.sum(grad, axis=i, keepdims=True)
        return grad

class sum(Operation):
    def __init__(self, x1, x2=None, *args, **kwargs):
        self.axis=None
        self.keepdims=False
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.sum(x1_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        if self.axis != None and self.keepdims == False:
            grad = np.broadcast_to(grad, self.x1.shape)
            self._apply_grad(self.x1, grad)
        else:
            self._apply_grad(self.x1, self._compress_grad(grad, self.x1))

class add(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.add(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self._apply_grad(self.x1, self._compress_grad(grad, x1_array))
        self._apply_grad(self.x2, self._compress_grad(grad, x2_array))


class subtract(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.subtract(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        self._apply_grad(self.x1, self._compress_grad(grad, self.x1))
        self._apply_grad(self.x2, -self._compress_grad(grad, self.x2))

class multiply(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.multiply(x1_array, x2_array, *args, **kwargs),requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self._apply_grad(self.x1, grad * x2_array)
        self._apply_grad(self.x2, grad * x1_array)

class dot(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args, **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.dot(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self._apply_grad(self.x1, np.dot(grad, x2_array.T))
        self._apply_grad(self.x2, np.dot(x1_array.T, grad))

class matmul(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.matmul(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self._apply_grad(self.x1, np.matmul(grad, x2_array.T))
        self._apply_grad(self.x2, np.matmul(x1_array.T, grad))

class divide(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.divide(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self._apply_grad(self.x1, self._compress_grad(grad / x2_array, self.x1))
        self._apply_grad(self.x2, self._compress_grad(-grad * x1_array / (x2_array ** 2), self.x2))

class floor_divide(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.floor_divide(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        return grad

class power(Operation):
    def __init__(self, x1, x2, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self.output = Tensor(np.power(x1_array, x2_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1, self.x2))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        x2_array = self._get_array(self.x2)
        self._apply_grad(self.x1, grad * x2_array * np.power(x1_array, x2_array - 1))
        self._apply_grad(self.x2, grad * np.power(x1_array, x2_array) * np.log(x1_array))

class exp(Operation):
    def __init__(self, x1, x2=None, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.exp(x1_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        self._apply_grad(self.x1, grad * np.exp(x1_array))

class log(Operation):
    def __init__(self, x1, x2=None, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.log(x1_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        self._apply_grad(self.x1, grad / x1_array)

class max(Operation):
    def __init__(self, x1, x2=None, axis=None, keepdims=False, *args, **kwargs):
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(x1, x2, axis, keepdims, *args,  **kwargs)

    def forward(self, axis, keepdims, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.max(x1_array, axis=axis, keepdims=keepdims, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        mask = (x1_array.max(axis=self.axis, keepdims=True) == x1_array).astype(int)
        self._apply_grad(self.x1, grad * mask)

class min(Operation):
    def __init__(self, x1, x2=None, axis=None, keepdims=False, *args, **kwargs):
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, axis, keepdims, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.min(x1_array, axis=axis, keepdims=keepdims, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        x1_array = self._get_array(self.x1)
        mask = (x1_array.min(axis=self.axis, keepdims=True) == x1_array).astype(int)
        self._apply_grad(self.x1, grad * mask)

class argmax(Operation):
    def __init__(self, x1, x2=None, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.argmax(x1_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        self._apply_grad(self.x1, grad)

class argmin(Operation):
    def __init__(self, x1, x2=None, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.argmin(x1_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        self._apply_grad(self.x1, grad)

class clip(Operation):
    def __init__(self, x1, x2=None, *args, **kwargs):
        super().__init__(x1, x2, *args,  **kwargs)

    def forward(self, *args, **kwargs):
        x1_array = self._get_array(self.x1)
        self.output = Tensor(np.clip(x1_array, *args, **kwargs), requires_grad=self._requires_grad(self.x1))
        return self.output

    def backward(self, grad):
        self._apply_grad(self.x1, grad)


from .preprocessing import *
from .NeuralNetworks import Sequential
from .LinearRegression import LinearRegression
from .layers import *
from .metrics import *
from .callbacks import *
from .optimizers import *
from .losses import *


import numpy as np
from tqdm import tqdm
from .layers import Linear

class LinearRegression:
    """
    A class for performing linear regression.

    Attributes
    ----------
    linear : Linear
        The linear layer used for regression.
    inputs : ndarray
        The input data for training.
    targets : ndarray
        The target data for training.
    prediction : ndarray
        The predicted output.
    loss : float
        The computed loss.

    Methods
    -------
    forward()
        Performs the forward pass to compute the output.
    predict(X)
        Generates predictions for the input data `X`.
    backward(lr)
        Performs the backward pass to update the parameters.
    fit(epochs, lr)
        Trains the model for a specified number of epochs.
    """
    def __init__(self, train_inputs, train_targets):
        """
        Initializes the LinearRegression object with training data.

        Parameters
        ----------
        train_inputs : ndarray
            The input data for training.
        train_targets : ndarray
            The target data for training.
        """
        self.linear = Linear(train_inputs.shape[1], train_targets.shape[1])
        self.inputs = self.linear.inputs = train_inputs
        self.targets = self.linear.targets = train_targets
        self.prediction = None
        self.loss = None

        if not isinstance(train_inputs, np.ndarray) or not isinstance(train_targets, np.ndarray):
            raise TypeError("Input data must be of type ndarray.")

    def forward(self):
        """
        Performs the forward pass to compute the output.

        Returns
        -------
        outputs : ndarray
            The output data.
        """
        self.outputs = self.linear.forward()
        return self.outputs

    def predict(self, X):
        """
        Generates predictions for the input data `X`.

        Parameters
        ----------
        X : ndarray
            The input data for which predictions are to be made.

        Returns
        -------
        predictions : ndarray
            The predicted output for the input data `X`.
        """
        self.linear.inputs = X
        return self.linear.forward()

    def backward(self, lr):
        """
        Performs the backward pass to update the parameters.

        Parameters
        ----------
        lr : float
            The learning rate for updating the parameters.
        """
        self.linear.backward(lr)

    def fit(self, epochs, lr):
        """
        Trains the model for a specified number of epochs.

        Parameters
        ----------
        epochs : int
            The number of epochs for training.
        lr : float
            The learning rate for updating the parameters.
        """
        for epoch in tqdm(range(epochs)):
            self.forward()
            self.backward(lr)




import numpy as np
from .layers import *
from .losses import *
from .metrics import *
from .optimizers import *
from .callbacks import *
from .preprocessing import *
from .tensor import *
from math import sqrt
from copy import deepcopy
import matplotlib.pyplot as plt
from tqdm import tqdm
import time
import pickle
import h5py
import sys
from datetime import datetime
import os

class Sequential:
    """
    A class for creating and training Feedforward Neural Networks (FNN) or Convolutional Neural Networks (CNN).

    Attributes
    ----------
    train_inputs : ndarray, Tensor
        Input data for training the model.
    train_input_batch : Tensor
        A batch of training inputs.
    train_targets : ndarray, Tensor
        Target data corresponding to training inputs.
    train_target_batch : Tensor
        A batch of training targets.
    train_outputs : Tensor
        Final output predictions during the training phase.
    train_output_batch : Tensor
        A batch of outputs during training.

    val_inputs : ndarray, Tensor
        Input data for validating the model (optional).
    val_input_batch : Tensor
        A batch of validation inputs.
val_targets : ndarray, Tensor
        Target data corresponding to validation inputs (optional).
    val_target_batch : Tensor
        A batch of validation targets.
    val_outputs : Tensor
        Final output predictions during the validation phase (not computed until further improvements).
    val_output_batch : Tensor
        A batch of outputs during validation.

    predictions : ndarray
        Predictions generated by the model for a given input array.

    train_layers : list
        List of layers used during training.
    val_layers : list
        List of layers used for validation.

    train_loss : float
        Calculated loss for the training dataset.
    val_loss : float
        Calculated loss for the validation dataset.
    train_loss_history : list
        History of training loss values for each epoch.
    val_loss_history : list
        History of validation loss values for each epoch.

    train_accuracy : float
        Accuracy computed on the training dataset.
    val_accuracy : float
        Accuracy computed on the validation dataset.
    train_accuracy_history : list
        History of training accuracy values for each epoch.
    val_accuracy_history : list
        History of validation accuracy values for each epoch.

    cost_function : Loss
        Cost function used for training.

    runtime : float
        Total time taken to train the model.

    optimizer : Optimizer
        Optimizer instance used to update model parameters.
    validation : bool
        Indicates whether a validation dataset is used.

    Methods
    -------
    add(layer : layer, activation : Activation)
        Adds a new layer to the model.
    forward()
        Performs the forward pass for both training and validation layers.
    backward()
        Executes the backward pass to compute gradients for training layers.
    verbose(verbose : int, epoch : int, epochs : int, start_time : float)
        Displays training and validation metrics based on the specified verbosity level.
    compile(cost_function : Loss, optimizer : Optimizer)
        Compiles the model with the specified cost function and optimizer.
    fit(epochs : int, batch_size : int, verbose : int, callbacks : list)
        Trains the model using the provided cost function, optimizer, and other parameters.
    predict(X : ndarray, Tensor)
        Generates predictions for the input data `X`.
    results()
        Visualizes loss and accuracy curves for both training and validation phases.
   save_params(name : str, extension : str)
        Saves the model's weights and biases to a file with the specified name and extension.
    load_params(path : str, parameters : dict)
        Loads model weights and biases from a file or dictionary.
    save_histories(name : str, extension : str)
        Saves the training and validation histories to a file with the specified name and extension.
    get_flatten_length()
        Computes the length of the flattened output from the `Flatten` layer.
    summary()
        Prints a summary of the model architecture.
    """

    def __init__(self, train_inputs, train_targets, val_inputs=None, val_targets=None):
        """
        Initializes the architecture with training and optional validation data.

        Parameters
        ----------
        train_inputs : ndarray, Tensor
            The input data used to train the model.
        train_targets : ndarray, Tensor
            The target labels or values corresponding to the train_inputs.
        val_inputs : ndarray, optional
            The input data used for validation. Default is None.
        val_targets : ndarray, optional
            The target labels or values corresponding to the `val_inputs`. Default is None.

        Raises
        ------
        TypeError
            If `train_inputs` and `train_targets` are not of type ndarray or Tensor.
            If `val_inputs` and `val_targets` are provided and not of type ndarray or Tensor.
        """
        self.train_inputs = Tensor(train_inputs) if isinstance(train_inputs, np.ndarray) else train_inputs
        self.train_input_batch = None
        self.train_targets = Tensor(train_targets) if isinstance(train_targets, np.ndarray) else train_targets
        self.train_target_batch = None
        self.train_outputs = None
        self.train_output_batch = None

        self.val_inputs = Tensor(val_inputs) if isinstance(val_inputs, np.ndarray) else val_inputs
        self.val_input_batch = None
        self.val_targets = Tensor(val_targets) if isinstance(val_targets, np.ndarray) else val_targets
        self.val_target_batch = None
        self.val_outputs = None
        self.val_output_batch = None

        self.batch_size = None
        self.val_batch_size = None

        self.predictions = None

        self.train_layers = []
        self.val_layers = []

        self.train_loss = 0.0
        self.train_loss_batch = 0.0
        self.val_loss = 0.0
        self.val_loss_batch = 0.0
        self.train_loss_history = []
        self.val_loss_history = []

        self.train_accuracy = 0.0
        self.val_accuracy = 0.0
        self.train_accuracy_history = []
        self.val_accuracy_history = []

        self.cost_function = None
        self.optimizer = None

        self.__is_trained__ = False
        self.__is_compiled__ = False
        self.runtime = 0.0
        self.validation = False

        if self.val_inputs is not None and self.val_targets is not None:
            self.validation = True

        if not isinstance(train_inputs, np.ndarray) and not isinstance(train_inputs, Tensor):
            raise TypeError('`train_inputs` must be of type ndarray or Tensor.')
        if not isinstance(train_targets, np.ndarray) and not isinstance(train_targets, Tensor):
            raise TypeError('`train_targets` must be of type ndarray or Tensor.')
        if self.val_inputs is not None and self.val_targets is not None:
            if not isinstance(val_inputs, np.ndarray) and not isinstance(val_inputs, Tensor):
                raise TypeError('`val_targets` must be of type ndarray or Tensor.')
            if not isinstance(val_targets, np.ndarray) and not isinstance(val_targets, Tensor):
                raise TypeError('`val_targets` must be of type ndarray or Tensor.')

    def get_flatten_length(self):
        """
        Computes the length of the flattened output from the `Flatten` layer.

        This method iterates through the training layers to find the `Flatten` layer and computes the length of its output.

        Returns
        -------
        int
            The length of the flattened output.

        Raises
        ------
        ValueError
            If there is no `Flatten` layer in the training layers.
        """
        self.train_layers[0].inputs = Tensor(self.train_inputs.array[0].reshape(1, *self.train_inputs.array[0].shape))
        for i in range(len(self.train_layers)):
            if isinstance(self.train_layers[i], Flatten):
                self.train_layers[i].forward()
                self.train_layers[0].inputs = self.train_inputs
                return self.train_layers[i].outputs.shape[1]
            elif i == len(self.train_layers) - 1:
                self.train_layers[0].inputs = self.train_inputs
                raise ValueError("There is no `Flatten` layer.")
            else:
                self.train_layers[i + 1].inputs = self.train_layers[i].forward()

    def add(self, layer):
        """
        Adds a layer to the training layers.

        Parameters
        ----------
        layer : Dense, Convolution2D, Pooling2D, Flatten or Dropout
            The layer to be added. Must be an instance of Dense, Convolution2D, Pooling2D, Flatten or Dropout.

        Raises
        ------
        TypeError
            If `layer` is not an instance of Dense, Convolution2D, Pooling2D, Flatten or Dropout.
        """
        if not isinstance(layer, (Dense, Convolution2D, Pooling2D, Flatten, Dropout)):
            raise TypeError("`layer` must be of type `Dense`, `Convolution2D`, `Pooling2D`, `Flatten` or `Dropout`.")

        self.train_layers.append(layer)

    def forward(self):
        """
        Performs the forward pass through the layers of the architecture for both
        training and validation.

        This method computes the output for both the training and validation data.
        During training, the input batch goes through the layers in sequence,
        and the output is stored in `train_output_batch`. If validation data is
        provided, the validation input batch goes through the validation layers
        and the output is stored in `val_output_batch`.

        Returns
        -------
        None
        """
        self.train_layers[0].inputs = self.train_input_batch
        if self.validation:
            self.val_layers[0].inputs = self.val_input_batch
        for i in range(len(self.train_layers)):
            if i + 1 == len(self.train_layers):
                self.train_output_batch = self.train_layers[i].forward()
                self.cost_function.forward(self.train_target_batch, self.train_output_batch)
            else:
                if isinstance(self.train_layers[i], Dropout):
                    self.train_layers[i].training = True
                self.train_layers[i + 1].inputs = self.train_layers[i].forward()
            if self.validation:
                if i + 1 == len(self.val_layers):
                    self.val_output_batch = self.val_layers[i].forward()
                else:
                    self.val_layers[i + 1].inputs = self.val_layers[i].forward()


    def predict(self, X):
        """
        Makes predictions for the given input `X` using the trained model.

        This method runs a forward pass with the provided input `X` and returns
        the predicted output based on the current state of the trained network.

        Be careful, you might encounter an error if the number of dimensions of `X`
        doesn't match the number of dimensions of `self.train_inputs`.

        Parameters
        ----------
        X : ndarray, Tensor
            The input data for which predictions are to be made.

        Returns
        -------
        predictions : ndarray
            The predicted output for the given input `X`.

        Raises
        ------
        TypeError
            If `X` is not of type ndarray or Tensor.
        """
        if not isinstance(X, np.ndarray) and not isinstance(X, Tensor):
            raise TypeError("`X` must be of type ndarray or Tensor.")
        if X.ndim != self.train_inputs.ndim:
            raise TypeError("`X`must have the same number of dimensions as `self.train_inputs`.")
        self.train_layers[0].inputs = Tensor(X) if isinstance(X, np.ndarray) else X
        for i in range(len(self.train_layers)):
            if isinstance(self.train_layers[i], Dropout):
                self.train_layers[i].training = False
            if i + 1 == len(self.train_layers):
                self.predictions = self.train_layers[i].forward()
                return self.predictions.array
            self.train_layers[i + 1].inputs = self.train_layers[i].forward()

    def backward(self):
        """
        Performs the backward pass through the layers to compute gradients.

        This method calculates the gradients of the loss function with respect to
        the parameters (weights and biases) of the network. The gradients are propagated backward through
        each layer.

        Returns
        -------
        None
        """
        self.dX = self.cost_function.backward()
        for layer in reversed(self.train_layers):
            self.dX = layer.backward(self.dX)

    def verbose(self, verbose, epoch, epochs, start_time):
        """
        Prints training and validation metrics during training at specified intervals.

        This method provides feedback during training based on the verbosity level
        chosen. It can print the loss and accuracy for both the training and validation
        datasets, and also the training time. The method adjusts the print frequency
        based on the `verbose` parameter.

        Parameters
        ----------
        verbose : int
            The verbosity level (0, 1, or 2).
            0 - No output.
            1 - Print at the end of the training.
            2 - Print at the end of each epoch.
        epoch : int
            The current epoch number.
        epochs : int
            The total number of epochs.
        start_time : float
            The starting time of training to calculate the runtime.

        Raises
        ------
        ValueError
            If the `verbose` parameter is not 0, 1, or 2.
        TypeError
            If `verbose`, `epochs`, or `start_time` are not of the correct type.

        Returns
        -------
        None
        """
        if not isinstance(verbose, int) and verbose is not None:
            raise TypeError('`verbose` must be of type int or None.')
        if not isinstance(epochs, int):
            raise TypeError('`epochs` must be of type int.')
        if not isinstance(start_time, float):
            raise TypeError('`start_time` must be of type float.')

        if verbose == 2:
            if epoch + 1 < epochs:
                if self.validation:
                    print(f"[TRAINING METRICS] train_loss: {np.around(self.train_loss, 5)} · "
                          f"train_accuracy: {np.around(self.train_accuracy, 5)}\n" +
                          f"[VALIDATION METRICS] val_loss: {np.around(self.val_loss, 5)} · "
                          f"val_accuracy: {np.around(self.val_accuracy, 5)}\n\n")
                else:
                    print(f"[TRAINING METRICS] train_loss: {np.around(self.train_loss, 5)} | " +
                          f"train_accuracy: {np.around(self.train_accuracy, 5)}\n\n")
            elif epoch + 1 == epochs:
                if self.validation:
                    self.runtime = time.time() - start_time
                    string1 = f"| [TRAINING METRICS] train_loss: {np.around(self.train_loss, 5)} · " + \
                              f"train_accuracy: {np.around(self.train_accuracy, 5)} |"
                    string2 = f"| [VALIDATION METRICS] val_loss: {np.around(self.val_loss, 5)} · " + \
                              f"val_accuracy: {np.around(self.val_accuracy, 5)} |"
                    string1_length = len(string1)
                    string2_length = len(string2)
                    print("\n" + string1_length * "-" + "\n" + string1 + "\n" +
                          string1_length * "-" + "\n" +
                          string2 + (string1_length - string2_length - 1) * " " + "\n" +
                          string1_length * "-")
                    print(f"{round(self.runtime, 5)} seconds")
                else:
                    self.runtime = time.time() - start_time
                    string = f"| [TRAINING METRICS] train_loss: {np.around(self.train_loss, 5)} · " + \
                             f"train_accuracy: {np.around(self.train_accuracy, 5)} |"
                    string_length = len(string)
                    print("\n" + string_length * "-" + "\n" + string + "\n" + string_length * "-")
                    print(f"{round(self.runtime, 5)} seconds")
        elif verbose == 1:
            if epoch + 1 == epochs:
                if self.validation:
                    string1 = f"| [TRAINING METRICS] train_loss: {np.around(self.train_loss, 5)} · " + \
                              f"train_accuracy: {np.around(self.train_accuracy, 5)} |"
                    string2 = f"| [VALIDATION METRICS] val_loss: {np.around(self.val_loss, 5)} · " + \
                              f"val_accuracy: {np.around(self.val_accuracy, 5)} |"
                    string1_length = len(string1)
                    string2_length = len(string2)
                    print("\n" + string1_length * "-" + "\n" + string1 + "\n" + string1_length * "-" + "\n" +
                          string2 + (string1_length - string2_length - 1) * " " + "\n" + string1_length * "-")
                else:
                    string = f"| [TRAINING METRICS] train_loss: {np.around(self.train_loss, 5)} · " + \
                             f"train_accuracy: {np.around(self.train_accuracy, 5)} |"
                    string_length = len(string)
                    print("\n" + string_length * "-" + "\n" + string + "\n" + string_length * "-")
        elif verbose == 0 or verbose is None:
            return
        else:
            raise ValueError("`verbose` must be 0, 1, or 2.")

    def compile(self, cost_function, optimizer=SGD(learning_rate=0.01)):
        """
        Compiles the model with the specified cost function and optimizer.

        Parameters
        ----------
        cost_function : Loss
            The cost function to be used for training.
        optimizer : Optimizer, optional
            The optimizer to be used for updating the model parameters. Default is SGD with a learning rate of 0.01.

        Raises
        ------
        ValueError
            If `cost_function` is not an instance of Loss.
            If `optimizer` is not an instance of Optimizer.

        Returns
        -------
        None
        """
        if not isinstance(cost_function, Loss):
            raise ValueError("`cost_function` must be of type `Loss`.")
        if not isinstance(optimizer, Optimizer):
            raise ValueError("`optimizer` must be of type `Optimizer`.")

        self.__is_compiled__ = True
        self.optimizer = optimizer
        self.cost_function = cost_function

    def fit(self, epochs=1000, batch_size=None, verbose=1, callbacks=[], get_output=False):
        """
        Trains the model using the provided cost function, optimizer, and other parameters.

        This method performs the training process for the neural network. It includes
        forward and backward passes, loss and accuracy computation and parameter updates
        using the chosen optimizer. During training, it also keeps track of the loss and
        accuracy for both the training and validation datasets (if validation data is provided).

        Parameters
        ----------
        epochs : int, optional
            The number of epochs for training. The default is 1000.
        batch_size : int, optional
            The number of samples per batch. If None, the entire dataset is used for each pass.
        verbose : int, optional
            The verbosity level for printing metrics during training. Default is 1.
        callbacks : list of Callback, optional
            A list of callback functions to extend training functionality. Each callback should be a callable
            object that implements the following methods:
            - `on_train_start(model)`: Called at the start of the training loop.
            - `on_epoch_start(model)`: Called at the start of each epoch.
            - `on_epoch_end(model)`: Called at the end of each epoch. If the callback is an instance
              of `LiveMetrics`, a `figure` parameter should be passed.
            - `on_train_end(model)`: Called at the end of the training loop.
            The default is an empty list.
        get_output : bool, optional
            If True the final output is computed. Default is True.

        Raises
        ------
        ValueError
            - If the model has not been compiled before fitting.
            - If `batch_size` is not of type `int` and is not `None`.
            - If `epochs` is not of type `int`.
            - If `verbose` is not of type `int` and is not `None`.
        TypeError
            - If `callbacks` is not a list of Callback objects.

        Returns
        -------
        None
        """
        if not self.__is_compiled__:
            raise ValueError("The model must be compiled before fitting.")
        if not isinstance(batch_size, int) and batch_size is not None:
            raise ValueError("`batch_size` must be of type `int` or `None`.")
        if not isinstance(epochs, int):
            raise ValueError("`epochs` must be of type `int`.")
        if not isinstance(verbose, int) and verbose is not None:
            raise ValueError("`verbose` must be of type `int` or `None`.")
        if not isinstance(callbacks, list) or not all(isinstance(callback, Callback) for callback in callbacks):
            raise TypeError("`callbacks` must be a list of Callback objects.")

        self.__is_trained__ = True

        start_time = time.time()

        if self.validation:
            self.val_layers = deepcopy(self.train_layers)

        self.batch_size = batch_size

        if self.batch_size is None:
            steps = 1
        else:
            steps = self.train_inputs.shape[0] // self.batch_size
            if steps * self.batch_size < self.train_inputs.shape[0]:
                steps += 1
            if self.validation:
                self.val_batch_size = self.val_inputs.shape[0] // steps
                if self.val_batch_size <= 0:
                    raise ValueError(
                        f"Validation batch size must be at least 1. (currently {self.val_batch_size}). "
                        f"Ensure validation input size is greater than {steps} or increase training batch size to "
                        f"{self.train_inputs.shape[0] // self.val_inputs.shape[0]}."
                    )

        if epochs > 1000:
            update = 25
        elif 1000 >= epochs > 100:
            update = 10
        elif epochs <= 100:
            update = 1

        loss = 0.0
        acc = 0.0

        tqdm_epochs = False
        tqdm_steps = False

        if verbose == 1:
            tqdm_epochs = True
        elif verbose is not None and verbose != 0 and verbose != 1:
            tqdm_steps = True

        for callback in callbacks:
            if isinstance(callback, LiveMetrics):
                figure = plt.figure()
            callback.on_train_start(self)

        for epoch in (epoch_bar := tqdm(range(epochs), disable=not tqdm_epochs, file=sys.stdout)):
            epoch_bar.set_description(f"Epoch [{epoch + 1}/{epochs}]")
            epoch_bar.set_postfix({"loss": loss, "accuracy": acc})

            for callback in callbacks:
                callback.on_epoch_start(self)

            train_accumulated_loss = 0
            train_accumulated_accuracy = 0

            val_accumulated_loss = 0
            val_accumulated_accuracy = 0

            for step in (step_bar := tqdm(range(steps), disable=not tqdm_steps, file=sys.stdout)):
                step_bar.set_description(f"Epoch [{epoch + 1}/{epochs}]")
                step_bar.set_postfix({"loss": loss, "accuracy": acc})

                for callback in callbacks:
                    callback.on_step_start(self)

                if self.batch_size is None:
                    self.train_input_batch = self.train_inputs
                    self.train_target_batch = self.train_targets

                    if self.validation:
                        self.val_input_batch = self.val_inputs
                        self.val_target_batch = self.val_targets
                else:
                    self.train_input_batch = Tensor(self.train_inputs.array[step * self.batch_size:(step + 1) * self.batch_size])
                    self.train_target_batch = Tensor(self.train_targets.array[step * self.batch_size:(step + 1) * self.batch_size])

                    if self.validation:
                        self.val_input_batch = Tensor(self.val_inputs.array[
                                               step * self.val_batch_size:(step + 1) * self.val_batch_size])
                        self.val_target_batch = Tensor(self.val_targets.array[
                                                step * self.val_batch_size:(step + 1) * self.val_batch_size])

                self.forward()

                for layer in self.train_layers:
                    layer.zero_grad()

                self.cost_function.zero_grad()

                self.backward()

                for i in range(len(self.train_layers)):
                    self.train_layers[i] = self.optimizer.update_params(self.train_layers[i])

                    if self.validation:
                        if isinstance(self.val_layers[i], Dense) or isinstance(self.val_layers[i], Convolution2D):
                            self.val_layers[i].weights = self.train_layers[i].weights
                            self.val_layers[i].biases = self.train_layers[i].biases

                loss = np.around(self.cost_function.output.array, 5)
                acc = accuracy(self.train_target_batch, self.train_output_batch)

                train_accumulated_loss += loss
                train_accumulated_accuracy += acc

                if self.validation:
                    val_accumulated_loss += self.cost_function.forward(self.val_target_batch, self.val_output_batch)
                    val_accumulated_accuracy += accuracy(self.val_target_batch, self.val_output_batch)

                for callback in callbacks:
                    callback.on_step_end(self)

            self.train_loss = train_accumulated_loss / steps
            self.train_accuracy = train_accumulated_accuracy / steps

            if epoch % update == 0 or epoch == 1:
                self.train_loss_history.append(self.train_loss)
                self.train_accuracy_history.append(self.train_accuracy)

            if self.validation:
                self.val_loss = (val_accumulated_loss / steps).array
                self.val_accuracy = val_accumulated_accuracy / steps

                if epoch % update == 0 or epoch == 1:
                    self.val_loss_history.append(self.val_loss)
                    self.val_accuracy_history.append(self.val_accuracy)

            self.verbose(verbose, epoch, epochs, start_time)

            for callback in callbacks:
                if isinstance(callback, LiveMetrics):
                    if epoch % update == 0 or epoch == 1:
                        callback.on_epoch_end(self, figure)
                else:
                    callback.on_epoch_end(self)

        if get_output:
            if self.batch_size is not None:
                self.train_outputs = Tensor(self.predict(self.train_inputs.array[:self.batch_size]))
                for step in range(1, steps):
                    self.train_outputs = Tensor(np.concatenate((self.train_outputs.array, Tensor(self.predict(
                        self.train_inputs.array[step * self.batch_size:(step + 1) * self.batch_size]))), axis=0))
            else:
                self.train_outputs = Tensor(self.predict(self.train_inputs))

        for callback in callbacks:
            callback.on_train_end(self)

    def results(self):
        """
        Plots the loss and accuracy evolution over epochs.

        This method generates plots to visualize the progress of the training
        process, showing how the loss and accuracy change for both training
        and validation datasets (if validation data is provided). This helps
        assess whether the model is learning and if it is overfitting.

        Raises
        ------
        RuntimeError
            If the model has not been trained yet.

        Returns
        -------
        None
        """
        if self.__is_trained__:
            figure, axs = plt.subplots(1, 2)

            axs[0].set_title("Loss Evolution")
            axs[0].set_xlabel("Epoch")
            axs[0].set_ylabel("Loss")
            axs[0].plot(self.train_loss_history, label="training dataset")
            if self.validation:
                axs[0].plot(self.val_loss_history, label="validation dataset")

            axs[1].set_title("Accuracy Evolution")
            axs[1].set_xlabel("Epoch")
            axs[1].set_ylabel("Accuracy")
            axs[1].plot(self.train_accuracy_history, label="training dataset")
            if self.validation:
                axs[1].plot(self.val_accuracy_history, label="validation dataset")

            axs[0].legend()
            axs[1].legend()

            plt.show()
        else:
            raise RuntimeError("The model has not been trained yet.")

    def save_params(self, name="parameters", extension="h5"):
        """
        Saves the trained model parameters (weights and biases) to a file.

        This method saves the weights and biases of all the layers of the trained model into a file with the specified name and extension.
        The parameters are saved with a timestamp appended to the base name.
        This allows the model to be restored later for further use or evaluation.

        Parameters
        ----------
        name : str, optional
            The base name of the file to save the parameters. Default is "parameters".
        extension : str, optional
            The file extension to use for saving the parameters. The default is "h5".

        Returns
        -------
        parameters : dict
            A dictionary containing the model's weights and biases. The dictionary includes the following keys:
            - "weights": List of weights for each layer.
            - "biases": List of biases for each layer.

        Raises
        ------
        RuntimeError
            If the model has not been trained yet.
        TypeError
            If `name` is not a string.
            If `extension` is not a string.
        ValueError
            If `extension` is not either 'h5' or 'pkl'.

        Notes
        -----
        The parameters are saved in a file with a timestamp appended to the base name. The file format can be either HDF5 (.h5) or pickle (.pkl).
        """
        if not isinstance(name, str):
            raise TypeError("`name` must be a string")
        if not isinstance(extension, str):
            raise TypeError("`extension` must be a string")
        if extension not in ("h5", "pkl"):
            raise ValueError("`extension` must be either 'h5' or 'pkl'")

        if self.__is_trained__:
            date_time = datetime.now().strftime("%m_%d_%Y-%H_%M_%S")
            parameters = {"weights": None, "biases": None, "weight_momentums": None, "bias_momentums": None, "weight_cache": None, "bias_cache": None}

            weights = []
            biases = []

            weight_momentums = []
            bias_momentums = []

            weight_cache = []
            bias_cache = []

            for layer in self.train_layers:
                if isinstance(layer, Dense) or isinstance(layer, Convolution2D):
                    weights.append(np.atleast_1d(layer.weights.array))
                    biases.append(np.atleast_1d(layer.biases.array))

                    weight_momentums.append(np.atleast_1d(layer.weight_momentums.array))
                    bias_momentums.append(np.atleast_1d(layer.bias_momentums.array))

                    weight_cache.append(np.atleast_1d(layer.weight_cache.array))
                    bias_cache.append(np.atleast_1d(layer.bias_cache.array))
                else:
                    weights.append(np.atleast_1d(0.0))
                    biases.append(np.atleast_1d(0.0))

                    weight_momentums.append(np.atleast_1d(0.0))
                    bias_momentums.append(np.atleast_1d(0.0))

                    weight_cache.append(np.atleast_1d(0.0))
                    bias_cache.append(np.atleast_1d(0.0))

            parameters["weights"] = weights
            parameters["biases"] = biases

            parameters["weight_momentums"] = weight_momentums
            parameters["bias_momentums"] = bias_momentums

            parameters["weight_cache"] = weight_cache
            parameters["bias_cache"] = bias_cache

            if extension == "pkl":
                with open(name + f"_{date_time}.pkl", 'wb') as f:
                    pickle.dump(parameters, f)

            elif extension == "h5":
                with h5py.File(name + f"_{date_time}.h5", 'w') as f:
                    weights_group = f.create_group("weights", track_order=True)
                    biases_group = f.create_group("biases", track_order=True)

                    weight_momentums_group = f.create_group("weight_momentums", track_order=True)
                    bias_momentums_group = f.create_group("bias_momentums", track_order=True)

                    weight_cache_group = f.create_group("weight_cache", track_order=True)
                    bias_cache_group = f.create_group("bias_cache", track_order=True)

                    for i in range(len(self.train_layers)):
                        weights_group.create_dataset(f"layer{i}", data=weights[i])
                        biases_group.create_dataset(f"layer{i}", data=biases[i])

                        weight_momentums_group.create_dataset(f"layer{i}", data=weight_momentums[i])
                        bias_momentums_group.create_dataset(f"layer{i}", data=bias_momentums[i])

                        weight_cache_group.create_dataset(f"layer{i}", data=weight_cache[i])
                        bias_cache_group.create_dataset(f"layer{i}", data=bias_cache[i])

            return parameters
        raise RuntimeError("The model has not been trained yet.")

    def load_params(self, path="parameters.h5", parameters=None):
        """
        Loads the model parameters (weights and biases) from a file or a given dictionary.

        This method loads the weights and biases from a file (either HDF5 or pickle format) or a dictionary
        and restores them to the corresponding layers in the model.
        This is useful for continuing training or evaluating a previously trained model.

        Parameters
        ----------
        path : str, optional
            The file path from which to load the parameters. Default is "parameters.h5".
        parameters : dict, optional
            A dictionary containing the model's weights and biases. Default is None.

        Raises
        ------
        TypeError
            If `path` is not a string.
            If `parameters` is not a dictionary.
        ValueError
            If the file extension is not either '.pkl' or '.h5'.

        Returns
        -------
        None

        Notes
        -----
        If `parameters` is provided, it will be used directly.
        Otherwise, the method will attempt to load the parameters from the file specified by `path`.
        The file format can be either HDF5 (.h5) or pickle (.pkl).
        """
        if not isinstance(path, str):
            raise TypeError("`path` must be a string")

        _, extension = os.path.splitext(path)
        if extension not in (".pkl", ".h5"):
            raise ValueError("`extension` must be either '.pkl' or '.h5'")

        if (type(parameters) == dict or parameters is None) and extension == ".pkl":
            if parameters is None:
                with open(path, 'rb') as f:
                    parameters = pickle.load(f)
            for i in range(len(self.train_layers)):
                if isinstance(self.train_layers[i], Dense) or isinstance(self.train_layers[i], Convolution2D):
                    self.train_layers[i].weights = Tensor(parameters["weights"][i])
                    self.train_layers[i].biases = Tensor(parameters["biases"][i])

                    self.train_layers[i].weight_momentums = Tensor(parameters["weight_momentums"][i])
                    self.train_layers[i].bias_momentums = Tensor(parameters["bias_momentums"][i])

                    self.train_layers[i].weight_cache = Tensor(parameters["weight_cache"][i])
                    self.train_layers[i].bias_cache = Tensor(parameters["bias_cache"][i])

        elif extension == ".h5" and parameters is None:
            with h5py.File(path, 'r') as f:
                weights = f["weights"]
                biases = f["biases"]

                weight_momentums = f["weight_momentums"]
                bias_momentums = f["bias_momentums"]

                weight_cache = f["weight_cache"]
                bias_cache = f["bias_cache"]

                for i in range(len(self.train_layers)):
                    if isinstance(self.train_layers[i], Dense) or isinstance(self.train_layers[i], Convolution2D):
                        self.train_layers[i].weights = Tensor(weights[f"layer{i}"].astype(np.float64)[:])
                        self.train_layers[i].biases = Tensor(biases[f"layer{i}"].astype(np.float64)[:])

                        self.train_layers[i].weight_momentums = Tensor(weight_momentums[f"layer{i}"].astype(np.float64)[:])
                        self.train_layers[i].bias_momentums = Tensor(bias_momentums[f"layer{i}"].astype(np.float64)[:])

                        self.train_layers[i].weight_cache = Tensor(weight_cache[f"layer{i}"].astype(np.float64)[:])
                        self.train_layers[i].bias_cache = Tensor(bias_cache[f"layer{i}"].astype(np.float64)[:])
        else:
            raise TypeError("`parameters` must be a dictionary")

    def save_histories(self, name="metrics_history", extension="h5"):
        """
        Saves the training and validation histories to a file.

        This method saves the training and validation histories (loss and accuracy) to a file with the specified name and extension.
        The histories are saved with a timestamp appended to the base name.

        Parameters
        ----------
        name : str, optional
            The base name of the file to save the histories. The default is "metrics_history".
        extension : str, optional
            The file extension to use for saving the histories. The default is "h5".

        Returns
        -------
        histories : dict
            A dictionary containing the training and validation histories. The dictionary includes the following keys:
            - "train_loss": List of training loss values.
            - "train_accuracy": List of training accuracy values.
            - "val_loss": List of validation loss values (if validation is enabled).
            - "val_accuracy": List of validation accuracy values (if validation is enabled).

        Raises
        ------
        TypeError
            If `name` is not a string.
            If `extension` is not a string.
        ValueError
            If `extension` is not either 'h5' or 'pkl'.

        Notes
        -----
        The histories are saved in a file with a timestamp appended to the base name. The file format can be either HDF5 (.h5) or pickle (.pkl).
        """
        if not isinstance(name, str):
            raise TypeError("`name` must be a string")
        if not isinstance(extension, str):
            raise TypeError("`extension` must be a string")
        if extension not in ("h5", "pkl"):
            raise ValueError("`extension` must be either 'h5' or 'pkl'")

        date_time = datetime.now().strftime("%m_%d_%Y-%H_%M_%S")
        histories = {
            "train_loss": self.train_loss_history,
            "train_accuracy": self.train_accuracy_history
        }
        if self.validation:
            histories["val_loss"] = self.val_loss_history
            histories["val_accuracy"] = self.val_accuracy_history

        if extension == "pkl":
            with open(name + f"_{date_time}.pkl", 'wb') as f:
                pickle.dump(histories, f)
        elif extension == "h5":
            with h5py.File(name + f"_{date_time}.h5", 'w') as f:
                f.create_dataset(f"train_loss", data=histories["train_loss"])
                f.create_dataset(f"train_accuracy", data=histories["train_accuracy"])
                if self.validation:
                    f.create_dataset(f"val_loss", data=histories["val_loss"])
                    f.create_dataset(f"val_accuracy", data=histories["val_accuracy"])

        return histories

    def summary(self):
        """
        Prints a summary of the model architecture.

        This method prints the shape of the output for each layer in the model.

        Returns
        -------
        None
        """
        params_count = 0

        self.predict(self.train_inputs.array[0].reshape(1, *self.train_inputs.array[0].shape))
        for i in range(len(self.train_layers)):
            if isinstance(self.train_layers[i], Dense) or isinstance(self.train_layers[i], Convolution2D):
                params_count += self.train_layers[i].weights.size
                if self.train_layers[i].biases is not None:
                    params_count += self.train_layers[i].biases.size
            print(f"{type(self.train_layers[i]).__name__}: {self.train_layers[i].outputs.shape}")

        print(f"\nNumber of parameters: {params_count}\n")


import numpy as np
from .tensor import *
from .layers import *

class Loss:
    """
    Base class for all loss functions.

    Methods
    -------
    forward(targets : Tensor, predictions : Tensor)
        Computes the loss.
    backward(targets : Tensor, predictions : Tensor)
        Computes the derivative of the loss.
    """
    def __init__(self):
        """
        Initializes the Loss class.
        """
        self.targets = None
        self.predictions = None
        self.output = None
        pass

    def forward(self, targets, predictions):
        """
        Computes the loss.

        Parameters
        ----------
        targets : Tensor
            Target data.
        predictions : Tensor
            Output data.

        Returns
        -------
        float
            The computed loss.
        """
        pass

    def backward(self):
        """
        Computes the derivative of the loss.

        Returns
        -------
        Tensor
            The derivative of the loss.
        """
        pass

    def zero_grad(self):
        pass

class MSE(Loss):
    """
    A class to perform the mean squared error loss.

    Methods
    -------
    forward(targets : Tensor, predictions : Tensor)
        Computes the mean squared error.
    """
    def __init__(self):
        """
        Initializes the MSE class.
        """
        super().__init__()

    def forward(self, targets, predictions):
        """
        Computes the mean squared error.

        Parameters
        ----------
        targets : Tensor
            Target data.
        predictions : Tensor
            Output data.

        Returns
        -------
        float
            The mean squared error.
        """
        self.targets = targets
        self.predictions = predictions

        diff = self.targets - self.predictions
        self.output = sum(diff * diff) / diff.size
        return self.output

    def backward(self):
        self.output.backward(1)
        return self.predictions.grad

    def zero_grad(self):
        self.output.zero_grad()

class BinaryCrossEntropy(Loss):
    """
    A class to perform binary cross entropy loss.

    Methods
    -------
    forward(targets : Tensor, predictions : Tensor)
        Computes the binary cross entropy loss.
    backward(targets : Tensor, predictions : Tensor)
        Computes the binary cross entropy derivative.
    """
    def __init__(self):
        """
        Initializes the BinaryCrossEntropy class.
        """
        super().__init__()

    def forward(self, targets, predictions):
        """
        Computes the binary cross entropy loss.

        Parameters
        ----------
        targets : Tensor
            Target data.
        predictions : Tensor
            Output data.

        Returns
        -------
        float
            The binary cross entropy loss.
        """
        self.targets = targets
        self.predictions = predictions

        e = 1e-10
        self.output = -(sum(self.targets * log(self.predictions + e) +
                        (1-self.targets) * log(1-self.predictions + e))) / len(self.targets)
        return self.output

    def backward(self):
        """
        Computes the binary cross entropy derivative.

        Returns
        -------
        Tensor
            The binary cross entropy derivative.
        """
        self.output.backward(1)
        return self.predictions.grad

    def zero_grad(self):
        self.output.zero_grad()


class CategoricalCrossEntropy(Loss):
    """
    A class to compute categorical cross-entropy loss and its derivative.

    Methods
    -------
    forward(targets : Tensor, predictions : Tensor)
        Computes the categorical cross-entropy loss.
    backward(targets : Tensor, predictions : Tensor)
        Computes the categorical cross-entropy derivative.
    """

    def __init__(self):
        """
        Initializes the CategoricalCrossEntropy class.
        """
        super().__init__()

    def forward(self, targets, predictions, from_logits=False):
        """
        Computes the categorical cross-entropy loss.

        Parameters
        ----------
        targets : Tensor
            Target data. Can be one-hot encoded or integer labels.
        predictions : Tensor
            Output probabilities (e.g., softmax predictions).

        Returns
        -------
        float
            The categorical cross-entropy loss.
        """
        self.targets = targets
        self.predictions = predictions

        self.predictions = clip(self.predictions, a_min=1e-15, a_max=1 - 1e-15)

        if from_logits:
            softmax = Softmax()
            softmax.inputs = self.predictions
            softmax.forward()

            loss = -sum(self.targets * log(softmax.outputs), axis=1)
        else:
            loss = -sum(self.targets * log(self.predictions), axis=1)

        self.output = sum(loss)/loss.size
        return self.output

    def backward(self):
        """
        Computes the categorical cross-entropy derivative.

        Returns
        -------
        Tensor
            Gradient of the loss with respect to predictions.
        """
        self.output.backward(1)
        return self.predictions.grad

    def zero_grad(self):
        self.output.zero_grad()

class NegativeLogLikelihood(Loss):
    """
    A class to compute negative log-likelihood loss and its derivative.

    Methods
    -------
    forward(targets : Tensor, predictions : Tensor)
        Computes the negative log-likelihood loss.
    backward(targets : Tensor, predictions : Tensor)
        Computes the negative log-likelihood derivative.
    """

    def __init__(self):
        """
        Initializes the CategoricalCrossEntropy class.
        """
        super().__init__()

    def forward(self, targets, predictions):
        """
        Computes the negative log-likelihood loss.

        Parameters
        ----------
        targets : Tensor
            Target data. Can be one-hot encoded or integer labels.
        predictions : Tensor
            Output probabilities (e.g., softmax predictions).

        Returns
        -------
        float
            The negative log-likelihood loss.
        """
        raise NotImplementedError

    def backward(self):
        """
        Computes the categorical cross-entropy derivative.

        Returns
        -------
        Tensor
            Gradient of the loss with respect to predictions.
        """
        raise NotImplementedError

    def zero_grad(self):
        raise NotImplementedError


import numpy as np
from .tensor import *

def accuracy(targets, predictions):
    """
    Computes the accuracy of the model's predictions.

    Parameters
    ----------
    targets : ndarray, Tensor
        The true target values.
    predictions : ndarray, Tensor
        The predicted output values.

    Returns
    -------
    float
        The accuracy of the model's predictions.

    Raises
    ------
    ValueError
        If the shapes of `targets` and `outputs` are incompatible.
    """
    if isinstance(targets, Tensor) or isinstance(targets, Operation):
        targets = targets.array

    if isinstance(predictions, Tensor) or isinstance(predictions, Operation):
        predictions = predictions.array

    if targets.shape[1] != 1 and predictions.shape[1] != 1:
        predicted_labels = np.argmax(predictions, axis=1)
        true_labels = np.argmax(targets, axis=1)
    else:
        predicted_labels = np.around(predictions)
        true_labels = np.around(targets)

    return  np.sum(predicted_labels == true_labels) / len(true_labels)

def mean_absolute_error(targets, predictions):
    """
    Prints the mean absolute error between the target values and the predicted output values.

    Parameters
    ----------
    targets : ndarray, Tensor
        The true target values.
    predictions : ndarray, Tensor
        The predicted output values.

    Raises
    ------
    ValueError
        If the lengths of `targets` and `outputs` are not the same.
    """
    if isinstance(targets, Tensor) or isinstance(targets, Operation):
        targets = targets.array

    if isinstance(predictions, Tensor) or isinstance(predictions, Operation):
        predictions = predictions.array

    return np.abs(targets - predictions).mean()


import numpy as np
import matplotlib.pyplot as plt
from .tensor import *

class Callback:
    """
    A base class for creating custom callbacks to extend the functionality of the training loop.

    Attributes
    ----------
    None

    Methods
    -------
    on_loop_start(model : Sequential, *args, **kwargs)
        Called at the start of the training loop.
    on_loop_end(model : Sequential, *args, **kwargs)
        Called at the end of the training loop.
    on_iteration_start(model : Sequential, *args, **kwargs)
        Called at the start of each epoch.
    on_iteration_end(model : Sequential, *args, **kwargs)
        Called at the end of each epoch.
    """

    def __init__(self):
        """
        Initializes the Callback object.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        pass

    def on_train_start(self, model, *args, **kwargs):
        """
        Called at the start of the training loop.

        This method can be overridden to perform actions at the beginning of the training loop.

        Parameters
        ----------
        model : Sequential
            The model being trained.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        None
        """
        pass

    def on_train_end(self, model, *args, **kwargs):
        """
        Called at the end of the training loop.

        This method can be overridden to perform actions at the end of the training loop.

        Parameters
        ----------
        model : Sequential
            The model being trained.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        None
        """
        pass

    def on_epoch_start(self, model, *args, **kwargs):
        """
        Called at the start of each epoch.

        This method can be overridden to perform actions at the beginning of each epoch.

        Parameters
        ----------
        model : Sequential
            The model being trained.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        None
        """
        pass

    def on_epoch_end(self, model, *args, **kwargs):
        """
        Called at the end of each epoch.

        This method can be overridden to perform actions at the end of each epoch.

        Parameters
        ----------
        model : Sequential
            The model being trained.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        None
        """
        pass

    def on_step_start(self, model, *args, **kwargs):
        """
        Called at the start of each step.

        Parameters
        ----------
        model : Sequential
            The model being trained.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        None
        """
        pass

    def on_step_end(self, model, *args, **kwargs):
        """
        Called at the end of each step.

        Parameters
        ----------
        model : Sequential
            The model being trained.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.
        Returns
        -------
        None
        """

class LiveMetrics(Callback):
    """
    A class for visualizing live metrics during the training of a neural network.

    Attributes
    ----------
    mode : int
        The mode of live metrics.
        1: Display loss and accuracy evolution.
        2: Display decision boundary.
        3: Display both loss/accuracy evolution and decision boundary.
    f1 : int
        The index of the first feature to use for plotting the decision boundary.
    f2 : int
        The index of the second feature to use for plotting the decision boundary.
    row_select : str
        Determines whether to use a limited or full set of rows for plotting.
        "limited": Use a limited number of rows.
        "full": Use the full set of rows.

    Methods
    -------
    on_iteration_end(model : Sequential, figure : Figure)
        Updates the live metrics plot based on the current state of the model.
    """

    def __init__(self, mode=3, f1=0, f2=1, row_select="limited"):
        """
        Initializes the LiveMetrics object with the specified parameters.

        Parameters
        ----------
        mode : int, optional
            The mode of live metrics to display. Default is 3.
            1: Display loss and accuracy evolution.
            2: Display decision boundary.
            3: Display both loss/accuracy evolution and decision boundary.
        f1 : int, optional
            The index of the first feature to use for plotting the decision boundary. Default is 0.
        f2 : int, optional
            The index of the second feature to use for plotting the decision boundary. Default is 1.
        row_select : str, optional
            Determines whether to use a limited or full set of rows for plotting. Default is "limited".
            "limited": Use a limited number of rows.
            "full": Use the full set of rows.

        Raises
        ------
        ValueError
            - If the `mode` parameter is not one of (1, 2, 3).
            - If the `row_select` parameter is not one of ("limited", "full").
            - If the `f1` or `f2` parameters are not integers.
        """
        self.mode = mode
        self.f1 = f1
        self.f2 = f2
        self.row_select = row_select

        if not isinstance(row_select, str):
            raise ValueError("`row_select` must be a string.")
        if not isinstance(f1, int):
            raise ValueError("`f1` must be an integer.")
        if not isinstance(f2, int):
            raise ValueError("`f2` must be an integer.")
        if not isinstance(mode, int):
            raise ValueError("`mode` must be an integer.")
        if self.mode not in (1, 2, 3):
            raise ValueError("`mode` must be one of (1, 2, 3).")
        if self.row_select not in ("limited", "full"):
            raise ValueError("`row_select` must be one of ('limited', 'full').")

    def on_epoch_end(self, model, figure):
        """
        Updates the live metrics plot based on the current state of the model.

        This method updates the live metrics plot based on the current state of the model.
        It supports different types of live metrics visualizations, including loss and accuracy
        evolution, and decision boundary plots.

        Parameters
        ----------
        model : Sequential
            The model whose metrics are to be visualized.
        figure : Figure
            The figure object to update with the live metrics plot.

        Raises
        ------
        ValueError
            If the `row_select` parameter is not one of ("limited", "full").
        """
        if self.mode == 1:
            figure.clear()

            plt.subplot(1, 2, 1)
            plt.title("Loss Evolution")
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.plot(model.train_loss_history, color="#1f77b4", label=f"training dataset: {round(float(model.train_loss_history[-1]), 3)}")
            if model.val_inputs is not None and model.val_targets is not None:
                plt.plot(model.val_loss_history, color="orange", label=f"validation dataset: {round(float(model.val_loss_history[-1]), 3)}")
            plt.legend()

            plt.subplot(1, 2, 2)
            plt.title("Accuracy Evolution")
            plt.xlabel("Epoch")
            plt.ylabel("Accuracy")
            plt.plot(model.train_accuracy_history, color="#1f77b4", label=f"training dataset: {round(float(model.train_accuracy_history[-1]), 3)}")
            if model.val_inputs is not None and model.val_targets is not None:
                plt.plot(model.val_accuracy_history, color="orange", label=f"validation dataset: {round(float(model.val_accuracy_history[-1]), 3)}")
            plt.legend()

            figure.canvas.draw()
            figure.canvas.flush_events()

        elif self.mode == 2:
            figure.clear()

            plt.subplot(1, 1, 1)
            if len(model.train_input_batch.array[0, :]) <= 2 and len(model.train_targets.array[0, :]) == 1:
                if len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "limited":
                    X_set, y_set = model.train_input_batch.array[:1000, :], model.train_targets.array[:1000, :]
                elif len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "full":
                    X_set, y_set = model.train_input_batch, model.train_targets
                elif (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "full") or (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "limited"):
                    X_set, y_set = model.train_input_batch.array, model.train_targets.array

                X1, X2 = np.meshgrid(np.arange(start=X_set[:, self.f1].min() - 1, stop=X_set[:, self.f1].max() + 1, step=0.01),
                                     np.arange(start=X_set[:, self.f2].min() - 1, stop=X_set[:, self.f2].max() + 1, step=0.01))
                plt.contourf(X1, X2, model.predict(Tensor(np.array([X1.ravel(), X2.ravel()]).T)).reshape(X1.shape),
                             alpha=0.3, cmap="coolwarm")

            if len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "limited":
                plt.scatter(model.train_input_batch.array[:1000, self.f1], model.train_input_batch.array[:1000, self.f2], c=model.train_output_batch.array[:1000, :], cmap="coolwarm", alpha=1)
            elif len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "full":
                plt.scatter(model.train_input_batch.array[:, self.f1], model.train_input_batch.array[:, self.f2], c=model.train_output_batch.array[:, :], cmap="coolwarm", alpha=1)
            elif (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "full") or (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "limited"):
                plt.scatter(model.train_input_batch.array[:, self.f1], model.train_input_batch.array[:, self.f2], c=model.train_output_batch[:, :], cmap="coolwarm", alpha=1)

            figure.canvas.draw()
            figure.canvas.flush_events()

        elif self.mode == 3:
            figure.clear()

            plt.subplot(1, 2, 2)
            if len(model.train_input_batch.array[0, :]) <= 2 and len(model.train_targets.array[0, :]) == 1:
                if len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "limited":
                    X_set, y_set = model.train_input_batch.array[:1000, :], model.train_targets.array[:1000, :]
                elif len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "full":
                    X_set, y_set = model.train_input_batch.array, model.train_targets.array
                elif (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "full") or (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "limited"):
                    X_set, y_set = model.train_input_batch.array, model.train_targets.array

                X1, X2 = np.meshgrid(np.arange(start=X_set[:, self.f1].min() - 1, stop=X_set[:, self.f1].max() + 1, step=0.01),
                                     np.arange(start=X_set[:, self.f2].min() - 1, stop=X_set[:, self.f2].max() + 1, step=0.01))
                plt.contourf(X1, X2, model.predict(Tensor(np.array([X1.ravel(), X2.ravel()]).T)).reshape(X1.shape),
                             alpha=0.3, cmap="coolwarm")

            if len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "limited":
                plt.scatter(model.train_input_batch.array[:1000, self.f1], model.train_input_batch.array[:1000, self.f2], c=model.train_output_batch.array[:1000, :], cmap="coolwarm", alpha=1)
            elif len(model.train_input_batch.array[:, 0]) > 1000 and self.row_select == "full":
                plt.scatter(model.train_input_batch.array[:, self.f1], model.train_input_batch.array[:, self.f2], c=model.train_output_batch.array[:, :], cmap="coolwarm", alpha=1)
            elif (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "full") or (len(model.train_input_batch.array[:, 0]) <= 1000 and self.row_select == "limited"):
                plt.scatter(model.train_input_batch.array[:, self.f1], model.train_input_batch.array[:, self.f2], c=model.train_output_batch.array[:, :], cmap="coolwarm", alpha=1)

            plt.subplot(2, 2, 1)
            plt.title("Loss Evolution")
            plt.ylabel("Loss")
            plt.plot(model.train_loss_history, color="#1f77b4", label=f"training dataset: {round(float(model.train_loss_history[-1]), 3)}")
            if model.val_inputs is not None and model.val_targets is not None:
                plt.plot(model.val_loss_history, color="orange", label=f"validation dataset: {round(float(model.val_loss_history[-1]), 3)}")
            plt.legend()

            plt.subplot(2, 2, 3)
            plt.title("Accuracy Evolution")
            plt.xlabel("Epoch")
            plt.ylabel("Accuracy")
            plt.plot(model.train_accuracy_history, color="#1f77b4", label=f"training dataset: {round(float(model.train_accuracy_history[-1]), 3)}")
            if model.val_inputs is not None and model.val_targets is not None:
                plt.plot(model.val_accuracy_history, color="orange", label=f"validation dataset: {round(float(model.val_accuracy_history[-1]), 3)}")
            plt.legend()

            figure.canvas.draw()
            figure.canvas.flush_events()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="melpy",
    version="1.1.0.1",
    author="Lenny Malard",
    author_email="lennymalard@gmail.com",
    description="Melpy is a package made to learn machine learning and make its uses as simple as possible.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[                      
        "numpy>=1.18.0",
        "matplotlib>=3.2.0",
        "tqdm>=4.50.0",
        "h5py>=3.12.0"
    ],                                             
    url="https://github.com/lennymalard/melpy-project",  
    packages=setuptools.find_packages(),
    classifiers=[                                 
        "Programming Language :: Python :: 3",   
        "License :: OSI Approved :: MIT License", 
        "Operating System :: OS Independent",  
        "Development Status :: 1 - Planning",
        "Natural Language :: French",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization"
    ],
)


import numpy as np
from melpy.tensor import *

def get_indices(image_shape, window_shape, stride):
    """
    Computes the indices for converting an image to a column matrix.

    Parameters
    ----------
    image_shape : tuple
        The shape of the input image (batch_size, channels, height, width).
    window_shape : int
        The size of the sliding window (kernel size).
    stride : int
        The stride of the sliding window.

    Returns
    -------
    k : ndarray
        Indices for the channel dimension.
    i : ndarray
        Indices for the height dimension.
    j : ndarray
        Indices for the width dimension.
    """
    output_height = int((image_shape[2] - window_shape + stride) // stride)
    output_width = int((image_shape[3] - window_shape + stride) // stride)

    level1 = np.repeat(np.arange(window_shape), window_shape)
    level1 = np.tile(level1, image_shape[1])

    increment = stride * np.repeat(np.arange(output_height), output_width)

    i = level1.reshape(-1, 1) + increment.reshape(1, -1)

    slide1 = np.tile(np.arange(window_shape), window_shape * image_shape[1])
    increment = stride * np.tile(np.arange(output_width), output_height)

    j = slide1.reshape(-1, 1) + increment.reshape(1, -1)

    k = np.repeat(np.arange(image_shape[1]), window_shape * window_shape).reshape(-1, 1)
    return k, i, j

def im2col(images, window_shape, stride):
    """
    Converts an image to a column matrix.

    Parameters
    ----------
    images : Tensor
        The input images with shape (batch_size, channels, height, width).
    window_shape : int
        The size of the sliding window (kernel size).
    stride : int
        The stride of the sliding window.

    Returns
    -------
    columns : Tensor
        The column matrix representation of the input images.
    """
    k, i, j = get_indices(images.shape, window_shape, stride)
    columns = Tensor(np.concatenate(images.array[:, k, i, j], axis=-1))
    return columns

def col2im(columns, image_shape, window_shape, stride):
    """
    Converts a column matrix back to an image.

    Parameters
    ----------
    columns : Tensor
        The column matrix representation of the images.
    image_shape : tuple
        The shape of the output image (batch_size, channels, height, width).
    window_shape : int
        The size of the sliding window (kernel size).
    stride : int
        The stride of the sliding window.

    Returns
    -------
    images : Tensor
        The reconstructed images with shape (batch_size, channels, height, width).
    """
    images = np.zeros(image_shape)
    k, i, j = get_indices(image_shape, window_shape, stride)
    cols_reshaped = np.array(np.hsplit(columns.array, image_shape[0]))
    np.add.at(images, (slice(None), k, i, j), cols_reshaped)
    return Tensor(images)
